from logging import getLogger
from re import search

from sqlalchemy import or_, distinct, desc, asc
from sqlalchemy.orm.exc import NoResultFound

from . import JournalDiary
from ..calculate.activity import ActivityCalculator
from ..calculate.power import PowerCalculator
from ..names import ACTIVE_DISTANCE, ACTIVE_TIME, ACTIVE_SPEED, MED_KM_TIME_ANY, MAX_MED_HR_M_ANY, CLIMB_ELEVATION, \
    CLIMB_DISTANCE, CLIMB_GRADIENT, CLIMB_TIME, TOTAL_CLIMB, MIN_KM_TIME_ANY, CALORIE_ESTIMATE, \
    ENERGY_ESTIMATE, MEAN_POWER_ESTIMATE, MAX_MEAN_PE_M_ANY, FITNESS_D_ANY, FATIGUE_D_ANY, _delta, M, S, \
    PERCENT_IN_Z_ANY, \
    KM, PC
from ..read.segment import SegmentReader
from ...data.climb import climbs_for_activity
from ...diary.database import summary_column
from ...diary.model import text, value, optional_text, from_field
from ...lib.date import format_seconds, time_to_local_time, to_time, HMS, local_date_to_time, to_date, MONTH, add_date, \
    time_to_local_date, YMD, YEAR, YM, DAY, HM, format_minutes, YMD_HM
from ...sql import ActivityGroup, ActivityJournal, StatisticJournal, StatisticName, ActivityTopic, ActivityTopicJournal, \
    ActivityTopicField

log = getLogger(__name__)


class ActivityDiary(JournalDiary):

    def __init__(self, *args, **kargs):
        super().__init__(*args, **kargs)

    def _read_journal_date(self, s, ajournal, date):
        yield text(self.__title(s, ajournal), tag='activity')
        yield from self.__read_activity_topics(s, ajournal, date)
        yield from self.__read_details(s, ajournal, date)

    def __read_activity_topics(self, s, ajournal, date):
        tjournal = ActivityTopicJournal.get_or_add(s, ajournal.file_hash)
        cache = tjournal.cache(s)
        # special case parentless fields
        for field in s.query(ActivityTopicField). \
                join(StatisticName). \
                filter(ActivityTopicField.activity_topic == None,
                       StatisticName.constraint == ajournal.activity_group). \
                order_by(ActivityTopicField.sort).all():
            yield from_field(field, cache[field])
        for topic in s.query(ActivityTopic). \
                filter(ActivityTopic.parent == None,
                       or_(ActivityTopic.activity_group_id == None,
                           ActivityTopic.activity_group_id == ajournal.activity_group.id)). \
                order_by(ActivityTopic.sort).all():
            yield list(self.__read_activity_topic(s, date, cache, topic))

    def __read_activity_topic(self, s, date, cache, topic):
        yield text(topic.name)
        if topic.description: yield text(topic.description)
        log.debug(f'topic id {topic.id}; fields {topic.fields}')
        for field in topic.fields:
            if field.statistic_name.constraint == str(topic.activity_group):
                yield from_field(field, cache[field])
        for child in topic.children:
            if child.activity_group == topic.activity_group and child.schedule.at_location(date):
                content = list(self.__read_activity_topic(s, date, cache, child))
                if content: yield content

    def __read_details(self, s, ajournal, date):
        zones = list(self.__read_zones(s, ajournal))
        if zones: yield [text('HR Zones (% time)')] + zones
        active_data = list(self.__read_active_data(s, ajournal, date))
        if active_data: yield [text('Activity Statistics')] + active_data
        climbs = list(self.__read_climbs(s, ajournal, date))
        if climbs: yield [text('Climbs')] + climbs
        for (title, template, re) in (('Min Time', MIN_KM_TIME_ANY, r'(\d+km)'),
                                      ('Med Time', MED_KM_TIME_ANY, r'(\d+km)'),
                                      ('Max Med Heart Rate', MAX_MED_HR_M_ANY, r'(\d+m)'),
                                      ('Max Mean Power Estimate', MAX_MEAN_PE_M_ANY, r'(\d+m)')):
            model = list(self.__read_template(s, ajournal, template, re, date))
            if model: yield [text(title)] + model

    @staticmethod
    def __read_zones(s, ajournal):
        percent_times = s.query(StatisticJournal).join(StatisticName). \
            filter(StatisticJournal.time == ajournal.start,
                   StatisticName.name.like(PERCENT_IN_Z_ANY),
                   StatisticName.owner == ActivityCalculator,
                   StatisticName.constraint == ajournal.activity_group) \
            .order_by(StatisticName.name).all()
        if percent_times:
            for zone, percent_time in reversed(list(enumerate((time.value for time in percent_times), start=1))):
                yield [value('Zone', zone, tag='hr-zone'), value('Percent time', percent_time)]

    @staticmethod
    def __title(s, ajournal):
        return '%s - %s  (%s) %s' % (time_to_local_time(to_time(ajournal.start), fmt=YMD_HM),
                                     time_to_local_time(to_time(ajournal.finish), fmt=HM),
                                     format_minutes((ajournal.finish - ajournal.start).seconds),
                                     ajournal.activity_group.name)

    @staticmethod
    def __read_active_data(s, ajournal, date):
        for name in (ACTIVE_DISTANCE, ACTIVE_TIME, ACTIVE_SPEED, MEAN_POWER_ESTIMATE):
            sjournal = StatisticJournal.at(s, ajournal.start, name, ActivityCalculator, ajournal.activity_group)
            if sjournal:
                yield value(sjournal.statistic_name.name, sjournal.value,
                            units=sjournal.statistic_name.units,
                            measures=sjournal.measures_as_model(date))
        for name in (_delta(FITNESS_D_ANY), _delta(FATIGUE_D_ANY)):
            for sjournal in StatisticJournal.at_like(s, ajournal.start, name, ActivityCalculator,
                                                     ajournal.activity_group):
                yield value(sjournal.statistic_name.name, sjournal.value,
                            units=sjournal.statistic_name.units,
                            measures=sjournal.measures_as_model(date))
        for name in (ENERGY_ESTIMATE, CALORIE_ESTIMATE):
            sjournal = StatisticJournal.at(s, ajournal.start, name, PowerCalculator, ajournal.activity_group)
            if sjournal:
                yield value(sjournal.statistic_name.name, sjournal.value,
                            units=sjournal.statistic_name.units,
                            measures=sjournal.measures_as_model(date))

    @staticmethod
    def __read_climbs(s, ajournal, date):
        total, climbs = climbs_for_activity(s, ajournal)
        if total:
            yield value('Total', total.value, measures=total.measures_as_model(date), units=M, tag='total-climb')
            for climb in climbs:
                yield [text('Climb'),
                       # todo - why are units hardcoded here?  already caused one bug when switched distance to km
                       value('Elevation', climb[CLIMB_ELEVATION].value, units=M,
                             measures=climb[CLIMB_ELEVATION].measures_as_model(date)),
                       value('Distance', climb[CLIMB_DISTANCE].value, units=KM),
                       value('Time', climb[CLIMB_TIME].value, units=S),
                       value('Gradient', climb[CLIMB_GRADIENT].value, units=PC)]

    def __read_template(self, s, ajournal, template, re, date):
        sjournals = s.query(StatisticJournal).join(StatisticName). \
            filter(StatisticJournal.time == ajournal.start,
                   StatisticName.name.like(template),
                   StatisticName.owner == ActivityCalculator,
                   StatisticName.constraint == ajournal.activity_group).order_by(StatisticName.name).all()
        for sjournal in self.__sort_journals(sjournals):
            if sjournal.value > 0:  # avoid zero power and anything else with silly value
                yield value(search(re, sjournal.statistic_name.name).group(1), sjournal.value,
                            units=sjournal.statistic_name.units, measures=sjournal.measures_as_model(date))

    @staticmethod
    def __sort_journals(sjournals):
        return sorted(sjournals,
                      # order by integer embedded in name
                      key=lambda sjournal: int(search(r'(\d+)', sjournal.statistic_name.name).group(1)))

    @staticmethod
    def __sort_names(statistic_names):
        return sorted(statistic_names,
                      # order by integer embedded in name
                      key=lambda statistic_name: int(search(r'(\d+)', statistic_name.name).group(1)))

    @optional_text('Activities', tag='activity')
    def _read_schedule(self, s, date, schedule):
        start, finish = local_date_to_time(schedule.start_of_frame(date)), local_date_to_time(schedule.next_frame(date))
        for group in s.query(ActivityGroup). \
                join(ActivityJournal, ActivityJournal.activity_group_id == ActivityGroup.id). \
                join(StatisticJournal, StatisticJournal.source_id == ActivityJournal.id). \
                join(StatisticName, StatisticJournal.statistic_name_id == StatisticName.id). \
                filter(StatisticName.name == ACTIVE_TIME,
                       StatisticJournal.time >= start,
                       StatisticJournal.time < finish).all():
            fields = list(self.__read_schedule_fields(s, date, schedule, group))
            if fields:
                yield [text(group.name)] + fields

    @staticmethod
    def __names(s, group, *names):
        for name in names:
            try:
                yield s.query(StatisticName). \
                    filter(StatisticName.name == name,
                           StatisticName.owner == ActivityCalculator,
                           StatisticName.constraint == group).one()
            except NoResultFound:
                log.warning(f'Missing "{name}" in database')

    @staticmethod
    def __names_like(s, group, name):
        return s.query(StatisticName). \
            filter(StatisticName.name.like(name),
                   StatisticName.owner == ActivityCalculator,
                   StatisticName.constraint == group).all()

    def __read_schedule_fields(self, s, start, schedule, group):
        for name in self.__names(s, group, ACTIVE_DISTANCE, ACTIVE_TIME, ACTIVE_SPEED,
                                 TOTAL_CLIMB, CLIMB_ELEVATION, CLIMB_DISTANCE, CLIMB_GRADIENT, CLIMB_TIME):
            column = list(summary_column(s, schedule, start, name))
            if column: yield column
        for name in self.__sort_names(self.__names_like(s, group, MIN_KM_TIME_ANY)):
            column = list(summary_column(s, schedule, start, name))
            if column: yield column
        for name in self.__sort_names(self.__names_like(s, group, MED_KM_TIME_ANY)):
            column = list(summary_column(s, schedule, start, name))
            if column: yield column
        for name in self.__sort_names(self.__names_like(s, group, MAX_MED_HR_M_ANY)):
            column = list(summary_column(s, schedule, start, name))
            if column: yield column


def active_dates(s, start, range, fmt):
    date_start = to_date(start)
    date_end = add_date(date_start, (1, range))
    time_start = local_date_to_time(date_start)
    time_end = local_date_to_time(date_end)
    times = s.query(distinct(ActivityJournal.start)). \
        filter(ActivityJournal.start >= time_start,
               ActivityJournal.start < time_end).all()
    return list(set(time_to_local_date(row[0]).strftime(fmt) for row in times))


def active_days(s, month):
    return active_dates(s, month, MONTH, YMD)


def active_months(s, year):
    return active_dates(s, year, YEAR, YM)


def activities_date(s, order, add_days=0):
    time = s.query(ActivityJournal.start). \
                order_by(order(ActivityJournal.start)).limit(1).one_or_none()
    if time:
        time = add_date(time_to_local_date(time[0]), (add_days, DAY))
        return time.strftime(YMD)
    else:
        return None


def activities_start(s):
    return activities_date(s, asc)


def activities_finish(s):
    return activities_date(s, desc, 1)


def activity_groups(s):
    return [row[0] for row in s.query(distinct(ActivityGroup.name)).all()]


def latest_activity(s):
    return s.query(ActivityJournal). \
        order_by(desc(ActivityJournal.start)).limit(1).one_or_none()


def activities_by_group(s):
    by_group = {}
    for group in s.query(ActivityGroup).all():
        activities = [time_to_local_time(activity.start)
                      for activity in s.query(ActivityJournal).filter(ActivityJournal.activity_group == group).all()]
        if activities:
            by_group[group.name] = activities
    return by_group




from abc import abstractmethod
from glob import iglob
from logging import getLogger
from os.path import join
from time import time

from ..pipeline import MultiProcPipeline, LoaderMixin
from ... import FatalException
from ...commands.args import base_system_path, PERMANENT
from ...fit.format.read import filtered_records
from ...lib.date import to_time
from ...lib.io import modified_file_scans
from ...lib.log import log_current_exception
from ...sql import Timestamp, Constant

log = getLogger(__name__)


class AbortImport(Exception):
    pass


class AbortImportButMarkScanned(AbortImport):
    pass


class FitReaderMixin(LoaderMixin):

    def __init__(self, *args, paths=None, sub_dir=None, **kargs):
        self.paths = paths
        self.sub_dir = sub_dir
        super().__init__(*args, **kargs)

    def _delete(self, s):
        # forcing does two things:
        # first, it by-passes the checks on last-scan date and duplicates
        # second, we delete any overlapping activities on load (when we know the times)
        # (without force, overlapping activities trigger an error)
        pass

    def _expand_paths(self, s, paths):
        from ...commands.upload import DOT_FIT
        if paths: return paths
        data_dir = base_system_path(self.base, version=PERMANENT)
        if self.sub_dir:
            data_dir = join(data_dir, self.sub_dir)
        else:
            log.warning('No sub_dir defined - will scan entire tree')
        return iglob(join(data_dir, '**/*' + DOT_FIT), recursive=True)

    def _missing(self, s):
        return modified_file_scans(s, self._expand_paths(s, self.paths), self.owner_out, self.force)

    def _run_one(self, s, file_scan):
        try:
            self._read(s, file_scan)
            file_scan.last_scan = time()
        except AbortImportButMarkScanned as e:
            log.warning(f'Could not process {file_scan} (scanned)')
            # log_current_exception()
            file_scan.last_scan = time()
        except FatalException:
            raise
        except Exception as e:
            log.warning(f'Could not process {file_scan} (ignored)')
            log_current_exception()

    def _read(self, s, file_scan):
        source, data = self._read_data(s, file_scan)
        s.commit()
        with Timestamp(owner=self.owner_out, source=source).on_success(s):
            loader = self._get_loader(s)
            self._load_data(s, loader, data)
            loader.load()
        return loader  # returned so coverage can be accessed

    @staticmethod
    def read_fit_file(data, *options):
        types, messages, records = filtered_records(data)
        return [record.as_dict(*options)
                for _, _, record in sorted(records,
                                           key=lambda r: r[2].timestamp if r[2].timestamp else to_time(0.0))]

    @staticmethod
    def _first(path, records, *names):
        return FitReaderMixin.assert_contained(path, records, names, 0)

    @staticmethod
    def _last(path, records, *names):
        return FitReaderMixin.assert_contained(path, records, names, -1)

    @staticmethod
    def assert_contained(path, records, names, index):
        try:
            return [record for record in records if record.name in names][index]
        except IndexError:
            msg = f'No {names} entry(s) in {path}'
            log.debug(msg)
            raise AbortImportButMarkScanned(msg)

    @abstractmethod
    def _read_data(self, s, file_scan):
        raise NotImplementedError()

    @abstractmethod
    def _load_data(self, s, loader, data):
        raise NotImplementedError()


def quote(text):
    return '"' + text + '"'


class MultiProcFitReader(FitReaderMixin, MultiProcPipeline):

    def _args(self, missing, start, finish):
        paths = ' '.join(quote(file_scan.path) for file_scan in missing[start:finish+1])
        log.info(f'Starting worker for {missing[start]} - {missing[finish]}')
        return paths




from bokeh.layouts import column
from bokeh.models import ColumnDataSource, Slider
from bokeh.plotting import figure
from bokeh.server.server import Server
from bokeh.themes import Theme

from bokeh.sampledata.sea_surface_temperature import sea_surface_temperature


def modify_doc(doc):
    df = sea_surface_temperature.copy()
    source = ColumnDataSource(data=df)

    plot = figure(x_axis_type='datetime', y_range=(0, 25), y_axis_label='Temperature (Celsius)',
                  title="Sea Surface Temperature at 43.18, -70.43")
    plot.line('time', 'temperature', source=source)

    def callback(attr, old, new):
        if new == 0:
            data = df
        else:
            data = df.rolling('{0}D'.format(new)).mean()
        source.data = ColumnDataSource(data=data).data

    slider = Slider(start=0, end=30, value=0, step=1, title="Smoothing by N Days")
    slider.on_change('value', callback)

    doc.add_root(column(slider, plot))

    doc.theme = Theme(filename="theme.yaml")


if __name__ == '__main__':

    # Setting num_procs here means we can't touch the IOLoop before now, we must
    # let Server handle that. If you need to explicitly handle IOLoops then you
    # will need to use the lower level BaseServer class.
    server = Server({'/': modify_doc})
    server.start()

    print('Opening Bokeh application on http://localhost:5006/')

    server.io_loop.add_callback(server.show, "/")
    server.io_loop.call_later(1, server.io_loop.stop)
    server.io_loop.start()




from ..commands.args import PATH, FORCE, parse_pairs, DEFINE, KARG, WORKER, KIT, BASE
from ..sql import PipelineType
from ..stats.pipeline import run_pipeline


def activities(args, sys, db):
    '''
## activities

    > ch2 activities [--kit ITEM [ITEM ...] --] [PATH [PATH ...]]

Read activities data from FIT files.

### Examples

    > ch2 activities -D kit=cotic ~/fit/2018-01-01.fit

will load the given file and associated the activity with the kit 'cotic'.

    > ch2 activities --kit

will load all unread files from the standard location
(constant Data.Dir, where they are stored by the upload command)
reading the kit from the file name (as encoded by the upload command).

Note: When using bash use `shopt -s globstar` to enable ** globbing.
    '''
    run_activity_pipelines(sys, db, args[BASE], paths=args[PATH], kit=args[KIT], force=args[FORCE],
                           worker=args[WORKER] is not None, id=args[WORKER],
                           define=parse_pairs(args[DEFINE], convert=False, comma=True), **parse_pairs(args[KARG]))


def run_activity_pipelines(sys, db, base, paths=tuple(), kit=True, force=False, worker=False, id=None, define=None,
                           **kargs):
    if define is None: define = {}
    run_pipeline(sys, db, base, PipelineType.ACTIVITY, paths=paths, force=force, worker=worker, id=id,
                 define=define, kit=kit, **kargs)




from argparse import ArgumentParser
from genericpath import exists
from logging import getLogger
from os import makedirs
from os.path import join
from re import sub
from typing import Mapping

from ..lib.date import to_date, to_time
from ..lib.utils import clean_path

log = getLogger(__name__)

# this can be modified during development.  it will be reset from setup.py on release.
CH2_VERSION = '0.31.6'
# new database on minor releases.  not sure this will always be a good idea.  we will see.
DB_VERSION = '-'.join(CH2_VERSION.split('.')[:2])
DB_EXTN = '.db'   # used to use .sql but auto-complete for sqlite3 didn't work

PERMANENT = 'permanent'

PROGNAME = 'ch2'
COMMAND = 'command'

ACTIVITIES = 'activities'
CONFIGURE = 'configure'
CONSTANTS = 'constants'
DIARY = 'diary'
DUMP = 'dump'
FIT = 'fit'
FIX_FIT = 'fix-fit'
GARMIN = 'garmin'
H, HELP = 'h', 'help'
JUPYTER = 'jupyter'
KIT = 'kit'
LOAD = 'load'
MONITOR = 'monitor'
NO_OP = 'no-op'
PACKAGE_FIT_PROFILE = 'package-fit-profile'
STATISTICS = 'statistics'
TEST_SCHEDULE = 'test-schedule'
UNLOCK = 'unlock'
UPGRADE = 'upgrade'
UPLOAD = 'upload'
WEB = 'web'

ACTIVITY = 'activity'
ACTIVITY_GROUP = 'activity-group'
ACTIVITY_GROUPS = 'activity-groups'
ACTIVITY_JOURNALS = 'activity-journals'
ACTIVITY_JOURNAL_ID = 'activity-journal-id'
ADD = 'add'
ADD_HEADER = 'add-header'
AFTER = 'after'
AFTER_BYTES = 'after-bytes'
AFTER_RECORDS = 'after-records'
ALL = 'all'
ALL_MESSAGES = 'all-messages'
ALL_FIELDS = 'all-fields'
ARG = 'arg'
BASE = 'base'
BIND = 'bind'
BORDER = 'border'
CHANGE = 'change'
CHECK = 'check'
CMD = 'cmd'
COMPACT = 'compact'
COMPONENT = 'component'
CONSTRAINT = 'constraint'
CONTEXT = 'context'
CSV = 'csv'
D = 'd'
DATA = 'data'
DATABASE = 'database'
DATE = 'date'
DEFAULT = 'default'
DEFINE = 'define'
DELETE = 'delete'
DESCRIBE = 'describe'
DESCRIPTION = 'description'
DEV = 'dev'
DIR = 'dir'
DISCARD = 'discard'
DROP = 'drop'
EMPTY = 'empty'
F = 'f'
FAST = 'fast'
FIELD = 'field'
FIELDS = 'fields'
FINISH = 'finish'
FIX_CHECKSUM = 'fix-checksum'
FIX_HEADER = 'fix-header'
FORMAT = 'format'
FORCE = 'force'
FTHR = 'fthr'
GREP = 'grep'
GROUP = 'group'
HEADER_SIZE = 'header-size'
HEIGHT = 'height'
INTERNAL = 'internal'
ITEM = 'item'
K = 'k'
KARG = 'karg'
LABEL = 'label'
LATITUDE = 'latitude'
LIKE = 'like'
LIMIT_BYTES = 'limit-bytes'
LIMIT_RECORDS = 'limit-records'
LOG = 'log'
LOGS = 'logs'
LONGITUDE = 'longitude'
LIST = 'list'
M, MESSAGE = 'm', 'message'
MATCH = 'match'
MAX_BACK_CNT = 'max-back-cnt'
MAX_COLUMNS = 'max-columns'
MAX_COLWIDTH = 'max-colwidth'
MAX_DROP_CNT = 'max-drop-cnt'
MAX_DELTA_T = 'max-delta-t'
MAX_FWD_LEN = 'max-fwd-len'
MAX_ROWS = 'max-rows'
MAX_RECORD_LEN = 'max-record-len'
MIN_SYNC_CNT = 'min-sync-cnt'
MODEL = 'model'
MONITOR_JOURNALS = 'monitor-journals'
MONTH = 'month'
MONTHS = 'months'
NAME = 'name'
NAME_BAD = 'name-bad'
NAME_GOOD = 'name-good'
NAMES = 'names'
NEW = 'new'
NOT = 'not'
NOTEBOOKS = 'notebooks'
O, OUTPUT = 'o', 'output'
OWNER = 'owner'
PASS = 'pass'
PATH = 'path'
P, PATTERN = 'p', 'pattern'
PLAN = 'plan'
PORT = 'port'
PRINT = 'print'
PROFILE = 'profile'
PROFILE_VERSION = 'profile-version'
PROTOCOL_VERSION = 'protocol-version'
PWD = 'pwd'
RAW = 'raw'
REBUILD = 'rebuild'
RECORDS = 'records'
REMOVE = 'remove'
REPLACE = 'replace'
RETIRE = 'retire'
ROOT = 'root'
RUN = 'run'
SEGMENT_JOURNALS = 'segment-journals'
SEGMENTS = 'segments'
SERVICE = 'service'
SET = 'set'
SCHEDULE = 'schedule'
SHOW = 'show'
SINGLE = 'single'
SLICES = 'slices'
SOURCE = 'source'
SOURCE_ID = 'source-id'
START = 'start'
STATISTIC_NAMES = 'statistic-names'
STATISTIC_JOURNALS = 'statistic-journals'
STATISTIC_QUARTILES = 'statistic-quartiles'
STATUS = 'status'
STOP = 'stop'
SUB_COMMAND = 'sub-command'
SYSTEM = 'system'
TABLE = 'table'
TABLES = 'tables'
TOKENS = 'tokens'
TOPIC = 'topic'
TUI = 'tui'
UNDO = 'undo'
UNLIKE = 'unlike'
UNSAFE = 'unsafe'
USER = 'user'
VALIDATE = 'validate'
V, VERBOSITY = 'v', 'verbosity'
VALUE = 'value'
VERSION = 'version'
W, WARN = 'w', 'warn'
WAYPOINTS = 'waypoints'
WIDTH = 'width'
WORKER = 'worker'
YEAR = 'year'
Y = 'y'


def mm(name): return '--' + name
def m(name): return '-' + name
def no(name): return 'no-%s' % name


class NamespaceWithVariables(Mapping):

    def __init__(self, ns):
        self._dict = vars(ns)

    def __getitem__(self, name):
        try:
            value = self._dict[name]
        except KeyError:
            value = self._dict[sub('-', '_', name)]
        return value

    def system_path(self, subdir=None, file=None, version=DB_VERSION, create=True):
        return base_system_path(self[BASE], subdir=subdir, file=file, version=version, create=create)

    def __iter__(self):
        return iter(self._dict)

    def __len__(self):
        return len(self.__dict__)

    def clone_with(self, **kargs):
        pass


def base_system_path(base, subdir=None, file=None, version=DB_VERSION, create=True):
    dir = base
    if version: dir = join(dir, version)
    if subdir: dir = join(dir, subdir)
    dir = clean_path(dir)
    if create and not exists(dir): makedirs(dir)
    if file:
        return join(dir, file)
    else:
        return dir


def make_parser():

    parser = ArgumentParser(prog=PROGNAME)

    parser.add_argument(mm(BASE), default=f'~/.ch2', metavar='DIR',
                        help='the base directory for data (default ~/.ch2)')
    parser.add_argument(mm(LOG), metavar='FILE',
                        help='the file name for the log (command name by default)')
    parser.add_argument(m(V), mm(VERBOSITY), default=4, type=int, metavar='VERBOSITY',
                        help='output level for stderr (0: silent; 5:noisy)')
    parser.add_argument(mm(TUI), action='store_true',
                        help='text user interface (no log to stdout)')
    parser.add_argument(mm(DEV), action='store_true', help='show stack trace on error')
    parser.add_argument(m(V.upper()), mm(VERSION), action='version', version=CH2_VERSION,
                        help='display version and exit')

    subparsers = parser.add_subparsers(title='commands', dest=COMMAND)

    # high-level commands used daily

    help = subparsers.add_parser(HELP, help='display help')
    help.add_argument(TOPIC, nargs='?', metavar=TOPIC,
                      help='the subject for help')

    web = subparsers.add_parser(WEB, help='the web interface (probably all you need)')
    web_cmds = web.add_subparsers(title='sub-commands', dest=SUB_COMMAND, required=True)

    def add_web_server_args(cmd):
        cmd.add_argument(mm(BIND), default='localhost', help='bind address (default localhost)')
        cmd.add_argument(mm(PORT), default=8000, type=int, help='port to use')

    add_web_server_args(web_cmds.add_parser(START, help='start the web server'))
    web_cmds.add_parser(STOP, help='stop the web server')
    web_cmds.add_parser(STATUS, help='display status of web server')
    add_web_server_args(web_cmds.add_parser(SERVICE, help='internal use only - use start/stop'))

    upload = subparsers.add_parser(UPLOAD, help='upload data (calls activities, monitor, statistics)')
    upload.add_argument(mm(FORCE), action='store_true', help='reprocess existing data')
    upload.add_argument(mm(KIT), m(K), action='append', default=[], metavar='ITEM',
                        help='kit items associated with activities')
    upload.add_argument(PATH, metavar='PATH', nargs='*', default=[], help='path to fit file(s) for activities')
    upload.add_argument(m(K.upper()), mm(KARG), action='append', default=[], metavar='NAME=VALUE',
                        help='keyword argument to be passed to the pipelines (can be repeated)')
    upload.add_argument(mm(FAST), action='store_true',
                        help='skip activity and statistics (just copy files)')
    upload.add_argument(mm(UNSAFE), action='store_true',
                        help='ignore duplicate files')
    upload.add_argument(mm(DELETE), action='store_true',
                        help='delete source on success')
    upload.add_argument(mm(REPLACE), action='store_true',
                        help='replace existing activity')

    diary = subparsers.add_parser(DIARY, help='daily diary and summary')
    diary.add_argument(DATE, metavar='DATE', nargs='?',
                       help='an optional date to display (default is today)')
    diary.add_argument(mm(FAST), action='store_true',
                       help='skip update of statistics on exit')
    diary_summary = diary.add_mutually_exclusive_group()
    diary_summary.add_argument(m(M), mm(MONTH), action='store_const', dest=SCHEDULE, const='m',
                               help='show monthly summary')
    diary_summary.add_argument(m(Y), mm(YEAR), action='store_const', dest=SCHEDULE, const='y',
                               help='show yearly summary')
    diary_summary.add_argument(mm(SCHEDULE), metavar='SCHEDULE',
                               help='show summary for given schedule')

    # low-level commands used often

    constant = subparsers.add_parser(CONSTANTS, help='set and examine constants')
    constant_cmds = constant.add_subparsers(title='sub-commands', dest=SUB_COMMAND, required=True)
    constant_show = constant_cmds.add_parser(SHOW, help='show a value (or all values)')
    constant_show.add_argument(NAME, nargs='?', metavar='NAME', help='name (omit for all)')
    constant_show.add_argument(DATE, nargs='?', metavar='DATE',
                               help='date of value to show (omit for all)')
    constant_add = constant_cmds.add_parser(ADD, help='add a new constant')
    constant_add.add_argument(NAME, metavar='NAME', help='name')
    constant_add.add_argument(CONSTRAINT, nargs='?', metavar='CONSTRAINT',
                              help='constraint (eg activity group)')
    constant_add.add_argument(mm(SINGLE), action='store_true', help='allow only a single (constant) value')
    constant_add.add_argument(mm(DESCRIPTION), help='optional description')
    constant_add.add_argument(mm(VALIDATE), help='optional validation class')
    constant_set = constant_cmds.add_parser(SET, help='set or modify a value')
    constant_set.add_argument(NAME, metavar='NAME', help='name')
    constant_set.add_argument(VALUE, metavar='VALUE', help='value')
    constant_set.add_argument(DATE, nargs='?', metavar='DATE',
                              help='date when measured (omit for all time)')
    constant_set.add_argument(mm(FORCE), action='store_true', help='allow over-writing existing values')
    constant_delete = constant_cmds.add_parser(DELETE, help='delete a value (or all values)')
    constant_delete.add_argument(NAME, metavar='NAME', help='name')
    constant_delete.add_argument(DATE, nargs='?', metavar='DATE',
                                 help='date of value to delete (omit for all)')
    constant_delete.add_argument(mm(FORCE), action='store_true', help='allow deletion of all values')
    constant_remove = constant_cmds.add_parser(REMOVE, help='remove a constant (after deleting all values)')
    constant_remove.add_argument(NAME, metavar='NAME', help='name')
    constant_remove.add_argument(mm(FORCE), action='store_true', help='allow remove of multiple constants')

    jupyter = subparsers.add_parser(JUPYTER, help='access jupyter')
    jupyter_cmds = jupyter.add_subparsers(title='sub-commands', dest=SUB_COMMAND, required=True)
    jupyter_cmds.add_parser(LIST, help='list available templates')
    jupyter_show = jupyter_cmds.add_parser(SHOW, help='display a template (starting server if needed)')
    jupyter_show.add_argument(NAME, help='the template name')
    jupyter_show.add_argument(ARG, nargs='*', help='template arguments')
    jupyter_cmds.add_parser(START, help='start a background service')
    jupyter_cmds.add_parser(STOP, help='stop the background service')
    jupyter_cmds.add_parser(STATUS, help='display status of background service')
    jupyter_cmds.add_parser(SERVICE, help='internal use only - use start/stop')

    kit = subparsers.add_parser(KIT, help='manage kit')
    kit_cmds = kit.add_subparsers(title='sub-commands', dest=SUB_COMMAND, required=True)
    kit_start = kit_cmds.add_parser(START, help='define a new item (new bike, new shoe)')
    kit_start.add_argument(GROUP, help='item group (bike, shoe, etc)')
    kit_start.add_argument(ITEM, help='item name (cotic, adidas, etc)')
    kit_start.add_argument(DATE, nargs='?', help='when created (default now)')
    kit_start.add_argument(mm(FORCE), action='store_true', help='confirm creation of a new group')
    kit_finish = kit_cmds.add_parser(FINISH, help='retire an item')
    kit_finish.add_argument(ITEM, help='item name')
    kit_finish.add_argument(DATE, nargs='?', help='when to retire (default now)')
    kit_finish.add_argument(mm(FORCE), action='store_true', help='confirm change of existing date')
    kit_delete = kit_cmds.add_parser(DELETE, help='remove all entries for an item or group')
    kit_delete.add_argument(NAME, help='item or group to delete')
    kit_delete.add_argument(mm(FORCE), action='store_true', help='confirm group deletion')
    kit_change = kit_cmds.add_parser(CHANGE, help='replace (or add) a part (wheel, innersole, etc)')
    kit_change.add_argument(ITEM, help='item name (cotic, adidas, etc)')
    kit_change.add_argument(COMPONENT, help='component type (chain, laces, etc)')
    kit_change.add_argument(MODEL, help='model description')
    kit_change.add_argument(DATE, nargs='?', help='when changed (default now)')
    kit_change.add_argument(mm(FORCE), action='store_true', help='confirm creation of a new component')
    kit_change.add_argument(mm(START), action='store_true', help='set default date to start of item')
    kit_undo = kit_cmds.add_parser(UNDO, help='remove a change')
    kit_undo.add_argument(ITEM, help='item name')
    kit_undo.add_argument(COMPONENT, help='component type')
    kit_undo.add_argument(MODEL, help='model description')
    kit_undo.add_argument(DATE, nargs='?', help='active date (to disambiguate models; default now)')
    kit_undo.add_argument(mm(ALL), action='store_true', help='remove all models (rather than single date)')
    kit_show = kit_cmds.add_parser(SHOW, help='display kit data')
    kit_show.add_argument(NAME, nargs='?', help='group or item to display (default all)')
    kit_show.add_argument(DATE, nargs='?', help='when to display (default now)')
    kit_show.add_argument(mm(CSV), action='store_true', help='CSV format')
    kit_statistics = kit_cmds.add_parser(STATISTICS, help='display statistics')
    kit_statistics.add_argument(NAME, nargs='?', help='group, item, component or model')
    kit_statistics.add_argument(mm(CSV), action='store_true', help='CSV format')
    kit_rebuild = kit_cmds.add_parser(REBUILD, help='rebuild database entries')
    kit_dump = kit_cmds.add_parser(DUMP, help='dump to script')
    kit_dump.add_argument(mm(CMD), help='command to use instead of ch2')

    # low-level commands use rarely

    configure = subparsers.add_parser(CONFIGURE, help='configure the database')
    configure_cmds = configure.add_subparsers(title='sub-commands', dest=SUB_COMMAND, required=True)
    configure_check = configure_cmds.add_parser(CHECK, help="check config")
    configure_check.add_argument(mm(no(DATA)), action='store_true', help='check database has no data loaded')
    configure_check.add_argument(mm(no(CONFIGURE)), action='store_true', help='check database has no configuration')
    configure_check.add_argument(mm(no(ACTIVITY_GROUPS)), action='store_true',
                              help='check database has no activity groups defined')
    configure_list = configure_cmds.add_parser(LIST, help='list available profiles')
    configure_load = configure_cmds.add_parser(LOAD, help="configure using the given profile")
    configure_profiles = configure_load.add_subparsers(title='profile', dest=PROFILE, required=True)
    from ..config.utils import profiles
    for name in profiles():
        configure_profile = configure_profiles.add_parser(name)
        configure_profile.add_argument(mm(no(DIARY)), action='store_true', help='skip diary creation (for migration)')
    configure_delete = configure_cmds.add_parser(DELETE, help='delete current data')
    configure_delete.add_argument(mm(FORCE), action='store_true', help='are you sure?')

    upgrade = subparsers.add_parser(UPGRADE, help='copy diary entries from a previous version')
    upgrade.add_argument(SOURCE, help='version or path to import')

    activities = subparsers.add_parser(ACTIVITIES, help='read activity data')
    activities.add_argument(mm(FORCE), action='store_true', help='re-read file and delete existing data')
    activities.add_argument(PATH, metavar='PATH', nargs='*', default=[], help='path to fit file(s)')
    activities.add_argument(mm(DEFINE), m(D.upper()), action='append', default=[], metavar='NAME=VALUE',
                            help='statistic to be stored with the activities (can be repeated)')
    activities.add_argument(m(K.upper()), mm(KARG), action='append', default=[], metavar='NAME=VALUE',
                            help='keyword argument to be passed to the pipelines (can be repeated)')
    activities.add_argument(mm(no(KIT)), action='store_false', dest=KIT, help='ignore kit encoded in file name')
    activities.add_argument(mm(WORKER), metavar='ID', type=int,
                            help='internal use only (identifies sub-process workers)')

    garmin = subparsers.add_parser(GARMIN, help='download monitor data from garmin connect')
    garmin.add_argument(DIR, metavar='DIR', nargs='?', help='the directory where FIT files are stored')
    garmin.add_argument(mm(USER), metavar='USER', help='garmin connect username')
    garmin.add_argument(mm(PASS), metavar='PASSWORD', help='garmin connect password')
    garmin.add_argument(mm(DATE), metavar='DATE', type=to_date, help='date to download')
    garmin.add_argument(mm(FORCE), action='store_true', help='allow longer date range')

    monitor = subparsers.add_parser(MONITOR, help='read monitor data')
    monitor.add_argument(mm(FORCE), action='store_true', help='re-read file and delete existing data')
    monitor.add_argument(PATH, metavar='PATH', nargs='*', help='path to fit file(s)')
    monitor.add_argument(m(K.upper()), mm(KARG), action='append', default=[], metavar='NAME=VALUE',
                         help='keyword argument to be passed to the pipelines (can be repeated)')
    monitor.add_argument(mm(WORKER), metavar='ID', type=int,
                         help='internal use only (identifies sub-process workers)')

    statistics = subparsers.add_parser(STATISTICS, help='(re-)generate statistics')
    statistics.add_argument(mm(FORCE), action='store_true',
                            help='delete existing statistics')
    statistics.add_argument(mm(LIKE), action='append', default=[], metavar='PATTERN',
                            help='run only matching pipeline classes')
    statistics.add_argument(mm(UNLIKE), action='append', default=[], metavar='PATTERN',
                            help='exclude matching pipeline classes')
    statistics.add_argument(START, metavar='START', nargs='?',
                            help='optional start date')
    statistics.add_argument(FINISH, metavar='FINISH', nargs='?',
                            help='optional finish date (if start also given)')
    statistics.add_argument(m(K.upper()), mm(KARG), action='append', default=[], metavar='NAME=VALUE',
                            help='keyword argument to be passed to the pipelines (can be repeated)')
    statistics.add_argument(mm(WORKER), metavar='ID', type=int,
                            help='internal use only (identifies sub-process workers)')

    dump = subparsers.add_parser(DUMP, help='display database contents')  # todo - this one needs tests!
    dump_format = dump.add_mutually_exclusive_group()
    dump_format.add_argument(mm(PRINT), action='store_const', dest=FORMAT, const=PRINT, help='default format')
    dump_format.add_argument(mm(CSV), action='store_const', dest=FORMAT, const=CSV, help='CVS format')
    dump_format.add_argument(mm(DESCRIBE), action='store_const', dest=FORMAT, const=DESCRIBE, help='summary format')
    dump.add_argument(mm(MAX_COLUMNS), metavar='N', type=int, help='pandas max_columns attribute')
    dump.add_argument(mm(MAX_COLWIDTH), metavar='N', type=int, help='pandas max_colwidth attribute')
    dump.add_argument(mm(MAX_ROWS), metavar='N', type=int, help='pandas max_rows attribute')
    dump.add_argument(mm(WIDTH), metavar='N', type=int, help='pandas width attribute')
    dump_sub = dump.add_subparsers(dest=SUB_COMMAND)
    dump_statistics = dump_sub.add_parser(STATISTICS)
    dump_statistics.add_argument(NAMES, nargs='*', metavar='NAME', help='statistic names')
    dump_statistics.add_argument(mm(START), metavar='TIME', help='start time')
    dump_statistics.add_argument(mm(FINISH), metavar='TIME', help='finish time')
    dump_statistics.add_argument(mm(OWNER), metavar='OWNER',
                                 help='typically the class that created the data')
    dump_statistics.add_argument(mm(CONSTRAINT), metavar='CONSTRAINT',
                                 help='a value that makes the name unique (eg activity group)')
    dump_statistics.add_argument(mm(SCHEDULE), metavar='SCHEDULE',
                                 help='the schedule on which some statistics are calculated')
    dump_statistics.add_argument(mm(SOURCE_ID), action='append', metavar='ID', type=int,
                                 help='the source ID for the statistic (can be repeated)')
    dump_statistic_quartiles = dump_sub.add_parser(STATISTIC_QUARTILES)
    dump_statistic_quartiles.add_argument(NAMES, nargs='*', metavar='NAME', help='statistic names')
    dump_statistic_quartiles.add_argument(mm(START), metavar='TIME', help='start time')
    dump_statistic_quartiles.add_argument(mm(FINISH), metavar='TIME', help='finish time')
    dump_statistic_quartiles.add_argument(mm(OWNER), metavar='OWNER',
                                          help='typically the class that created the data')
    dump_statistic_quartiles.add_argument(mm(CONSTRAINT), metavar='CONSTRAINT',
                                          help='a value that makes the name unique (eg activity group)')
    dump_statistic_quartiles.add_argument(mm(SCHEDULE), metavar='SCHEDULE',
                                          help='the schedule on which some statistics are calculated')
    dump_statistic_quartiles.add_argument(mm(SOURCE_ID), action='append', metavar='ID', type=int,
                                          help='the source ID for the statistic (can be repeated)')
    dump_table = dump_sub.add_parser(TABLE)
    dump_table.add_argument(NAME, metavar='NAME', help='table name')
    dump.set_defaults(format=PRINT)

    fit = subparsers.add_parser(FIT, help='display contents of fit file')
    fit_cmds = fit.add_subparsers(title='sub-commands', dest=SUB_COMMAND, required=True)
    fit_grep = fit_cmds.add_parser(GREP, help='show matching entries')
    fit_records = fit_cmds.add_parser(RECORDS, help='show high-level structure (ordered by time)')
    fit_tables = fit_cmds.add_parser(TABLES, help='show high-level structure (grouped in tables)')
    fit_csv = fit_cmds.add_parser(CSV, help='show high-level structure (in CSV format)')
    fit_tokens = fit_cmds.add_parser(TOKENS, help='show low-level tokens')
    fit_fields = fit_cmds.add_parser(FIELDS, help='show low-level fields (within tokens)')

    def add_fit_general(cmd):
        cmd.add_argument(mm(AFTER_RECORDS), type=int, metavar='N', default=None,
                         help='skip initial records')
        cmd.add_argument(mm(LIMIT_RECORDS), type=int, metavar='N', default=-1,
                         help='limit number of records displayed')
        cmd.add_argument(mm(AFTER_BYTES), type=int, metavar='N', default=None,
                         help='skip initial bytes')
        cmd.add_argument(mm(LIMIT_BYTES), type=int, metavar='N', default=-1,
                         help='limit number of bytes displayed')
        cmd.add_argument(m(W), mm(WARN), action='store_true',
                         help='log additional warnings')
        cmd.add_argument(mm(no(VALIDATE)), action='store_true',
                         help='do not validate checksum, length')
        cmd.add_argument(mm(MAX_DELTA_T), type=float, metavar='S',
                         help='validate seconds between timestamps (and non-decreasing)')
        cmd.add_argument(mm(NAME), action='store_true',
                         help='print file name')
        cmd.add_argument(PATH, metavar='PATH', nargs='+',
                         help='path to fit file')

    def add_fit_grep(cmd):
        cmd.add_argument(mm(NOT), action='store_true',
                         help='print file names that don\'t match (with --name)')
        cmd.add_argument(mm(MATCH), type=int, default=-1,
                         help='max number of matches (default -1 for all)')
        cmd.add_argument(mm(COMPACT), action='store_true',
                         help='no space between records')
        cmd.add_argument(mm(CONTEXT), action='store_true',
                         help='display entire record')
        cmd.add_argument(m(P), mm(PATTERN), nargs='+', metavar='MSG:FLD[=VAL]', required=True,
                         help='pattern to match (separate from PATH with --)')

    def add_fit_high_level(cmd):
        cmd.add_argument(m(M), mm(MESSAGE), nargs='+', metavar='MSG',
                         help='display named messages')
        cmd.add_argument(m(F), mm(FIELD), nargs='+', metavar='FLD',
                         help='display named fields')
        cmd.add_argument(mm(INTERNAL), action='store_true',
                         help='display internal messages')

    def add_fit_very_high_level(cmd):
        cmd.add_argument(mm(ALL_MESSAGES), action='store_true',
                         help='display undocumented messages')
        cmd.add_argument(mm(ALL_FIELDS), action='store_true',
                         help='display undocumented fields')

    for cmd in fit_grep, fit_records, fit_tables, fit_csv, fit_tokens, fit_fields:
        add_fit_general(cmd)

    for cmd in fit_records, fit_tables, fit_csv:
        add_fit_high_level(cmd)

    for cmd in fit_records, fit_tables:
        add_fit_very_high_level(cmd)

    add_fit_grep(fit_grep)

    for cmd in fit_grep, fit_records, fit_tables:
        cmd.add_argument(mm(WIDTH), type=int,
                         help='display width')

    fix_fit = subparsers.add_parser(FIX_FIT, help='fix a corrupted fit file')
    fix_fit.add_argument(PATH, metavar='PATH', nargs='+',
                         help='path to fit file')
    fix_fit.add_argument(m(W), mm(WARN), action='store_true',
                         help='additional warning messages')
    fix_fit_output = fix_fit.add_argument_group(title='output (default hex to stdout)').add_mutually_exclusive_group()
    fix_fit_output.add_argument(m(O), mm(OUTPUT), action='store',
                                help='output file for fixed data (otherwise, stdout)')
    fix_fit_output.add_argument(mm(DISCARD), action='store_true',
                                help='discard output (otherwise, stdout)')
    fix_fit_output.add_argument(mm(RAW), action='store_true',
                                help='raw binary to stdout (otherwise, hex encoded)')
    fix_fit_output.add_argument(mm(NAME_BAD), action='store_false', dest=NAME, default=None,
                                help='print file name if bad')
    fix_fit_output.add_argument(mm(NAME_GOOD), action='store_true', dest=NAME, default=None,
                                help='print file name if good')
    fix_fit_process = fix_fit.add_argument_group(title='processing (default disabled)')
    fix_fit_process.add_argument(mm(ADD_HEADER), action='store_true',
                                 help='preprend a new header')
    fix_fit_stage = fix_fit_process.add_mutually_exclusive_group()
    fix_fit_stage.add_argument(mm(DROP), action='store_true',
                               help='search for data that can be dropped to give a successful parse')
    fix_fit_stage.add_argument(mm(SLICES), metavar='A:B,C:D,...',
                               help='data slices to pick')
    fix_fit_stage.add_argument(mm(START), type=to_time, metavar='TIME',
                               help='change start time')
    fix_fit_process.add_argument(mm(FIX_HEADER), action='store_true',
                                 help='modify the header')
    fix_fit_process.add_argument(mm(FIX_CHECKSUM), action='store_true',
                                 help='modify the checksum')
    fix_fit_process.add_argument(mm(no(FORCE)), action='store_false', dest=FORCE,
                                 help='don\'t parse record contents')
    fix_fit_process.add_argument(mm(no(VALIDATE)), action='store_false', dest=VALIDATE,
                                 help='don\'t validate the final data')
    fix_fit_params = fix_fit.add_argument_group(title='parameters')
    fix_fit_params.add_argument(mm(HEADER_SIZE), type=int, metavar='N',
                                help='header size (validation and/or new header)')
    fix_fit_params.add_argument(mm(PROTOCOL_VERSION), type=int, metavar='N',
                                help='protocol version (validation and/or new header)')
    fix_fit_params.add_argument(mm(PROFILE_VERSION), type=int, metavar='N',
                                help='profile version (validation and/or new header)')
    fix_fit_params.add_argument(mm(MIN_SYNC_CNT), type=int, metavar='N', default=3,
                                help='minimum number of records to read for synchronization')
    fix_fit_params.add_argument(mm(MAX_RECORD_LEN), type=int, metavar='N', default=None,
                                help='maximum record length')
    fix_fit_params.add_argument(mm(MAX_DROP_CNT), type=int, metavar='N', default=1,
                                help='maximum number of gaps to drop')
    fix_fit_params.add_argument(mm(MAX_BACK_CNT), type=int, metavar='N', default=3,
                                help='maximum number of readable records to discard in a single gap')
    fix_fit_params.add_argument(mm(MAX_FWD_LEN), type=int, metavar='N', default=200,
                                help='maximum number of bytes to drop in a single gap')
    fix_fit_params.add_argument(mm(MAX_DELTA_T), type=float, metavar='S',
                                help='max number of seconds between timestamps')

    noop = subparsers.add_parser(NO_OP,
                                 help='used within jupyter (no-op from cmd line)')

    package_fit_profile = subparsers.add_parser(PACKAGE_FIT_PROFILE,
                                                help='parse and save the global fit profile (dev only)')
    package_fit_profile.add_argument(PATH, metavar='PROFILE',
                                     help='the path to the profile (Profile.xlsx)')
    package_fit_profile.add_argument(m(W), mm(WARN), action='store_true',
                                     help='additional warning messages')

    test_schedule = subparsers.add_parser(TEST_SCHEDULE, help='print schedule locations in a calendar')
    test_schedule.add_argument(SCHEDULE, metavar='SCHEDULE',
                               help='schedule to test')
    test_schedule.add_argument(mm(START), metavar='DATE',
                               help='date to start displaying data')
    test_schedule.add_argument(mm(MONTHS), metavar='N', type=int,
                               help='number of months to display')

    unlock = subparsers.add_parser(UNLOCK, help='remove database locking')

    return parser


def bootstrap_dir(base, *args, configurator=None, post_config=None):

    from ..lib.log import make_log_from_args
    from ..sql.database import Database, connect
    from ..sql.system import System

    args = [mm(BASE), base] + list(args)
    if configurator:
        ns, db = connect(args)
        sys = System(ns)
        with db.session_context() as s:
            configurator(sys, s, base)
    args += post_config if post_config else []
    ns = NamespaceWithVariables(make_parser().parse_args(args))
    make_log_from_args(ns)
    db = Database(ns)
    sys = System(ns)

    return ns, sys, db


def parse_pairs(pairs, convert=True, multi=False, comma=False):
    # simple name, value pairs. owner and constraint supplied by command.
    d = {}
    if pairs is not None:
        for pair in pairs:
            name, value = pair.split('=', 1)
            if convert:
                for type in (int, float, to_time):
                    try:
                        value = type(value)
                        break
                    except ValueError:
                        pass
            if multi:
                if name not in d:
                    d[name] = []
                d[name].append(value)
            elif comma:
                if name in d:
                    d[name] = d[name] + ',' + value
                else:
                    d[name] = value
            else:
                d[name] = value
    return d




from urwid import FIXED

from .focus import FocusWrap


class Fixed(FocusWrap):
    """
    Convert a widget to fixed size.  The height is whatever the widget
    flows to for the given width.
    """

    _sizing = frozenset([FIXED])

    def __init__(self, w, width):
        super().__init__(w)
        self._size = w.pack((width,))

    def pack(self, size, focus=False):
        return self._size

    def render(self, size, focus=False):
        if size != ():
            raise Exception('Using fixed widget incorrectly (received size of %s)' % size)
        return super().render((self._size[0], ), focus)





def to_tree(model, format, children):
    '''
    this will yield lines formatted as a tree.

    model is the data to be displayed (for example, nested dicts).

    format is a function that takes the current model and returns a label plus an optional list of children

    children takes the current model and returns children from sub-nodes.  these are combined with the children
    return by format.
    '''
    line, extra = format(model)
    yield line
    all_children = (list(extra) if extra else []) + list(children(model))
    for child in all_children:
        last = child == all_children[-1]
        prefix = '`-' if last else '+-'
        for line in to_tree(child, format, children):
            yield prefix + line
            prefix = '  ' if last else '| '


def to_csv(model, format, children):
    '''
    same interface as above, but returns csv (kinda).
    '''
    parent_line, extra = format(model)
    all_children = (list(extra) if extra else []) + list(children(model))
    displayed = False
    for child in all_children:
        for child_line in to_csv(child, format, children):
            yield parent_line + ',' + child_line
            displayed = True
    if not displayed:
        yield parent_line




from logging import getLogger

import rasterio as rio
import numpy as np
from pyproj import Proj, transform
from rasterio.enums import ColorInterp
from rasterio.plot import reshape_as_image
from scipy.ndimage import shift
from skimage.draw import polygon_perimeter

from ch2.stats.names import LATITUDE, LONGITUDE

log = getLogger(__name__)

RGB = 'RGB'
RGB_INDICES = (ColorInterp.red, ColorInterp.green, ColorInterp.blue)
GTIFF = 'GTiff'

COUNT = 'count'
CRS = 'crs'
DRIVER = 'driver'
DTYPES = 'dtypes'
HEIGHT = 'height'
PHOTOMETRIC = 'photometric'
TRANSFORM = 'transform'
WIDTH = 'width'

LATLONG = 'latlong'
WGS84 = 'WGS84'


def create_image(template, data, transform=None):
    log.debug('Creating new image')
    transform = transform or template.transform
    profile = template.profile.copy()
    profile.update({DRIVER: GTIFF,
                    COUNT: data.shape[0],
                    HEIGHT: data.shape[1],
                    WIDTH: data.shape[2],
                    TRANSFORM: transform,
                    PHOTOMETRIC: RGB})
    new_image = rio.MemoryFile().open(**profile)
    new_image.write(data)
    return new_image


def write_image(image, path):
    with rio.open(path, 'w', **image.profile) as dest:
        dest.write(image.read())


def read_image_rgb(path):
    image = rio.open(path, 'r+')
    image.colorinterp = RGB
    return image


def extract_rgb(image):
    color_map = dict(zip(image.colorinterp, image.indexes))
    rgb_indices = [color_map[ci] for ci in RGB_INDICES]
    return image.read(rgb_indices, masked=True)


def matplot_image(ax, image):
    # based on rasterio.plot, but accepts in-memory images
    rgb = extract_rgb(image)
    rgb = reshape_as_image(rgb)
    # https://stackoverflow.com/questions/24739769/matplotlib-imshow-plots-different-if-using-colormap-or-rgb-array
    lo, hi = np.min(rgb), np.max(rgb)
    rgb = ((rgb - lo) / (hi - lo)) ** (1 / 2.2)
    ax.imshow(rgb)


def matplot_route(ax, image, df):
    lat, lon = df[LATITUDE].values, df[LONGITUDE].values
    x, y = latlon_to_xy(image, lat, lon)
    ax.plot(x, y, color='red', linewidth='1')


def latlon_to_xy(image, lat, lon):
    # https://gis.stackexchange.com/a/129857
    p1 = Proj(proj=LATLONG, datum=WGS84)
    p2 = Proj(image.crs)
    east, north = transform(p1, p2, lon, lat)
    row, col = image.index(east, north)
    return col, row


def xy_to_latlon(image, x, y):
    east, north = image.xy(x, y)
    p1 = Proj(image.crs)
    p2 = Proj(proj=LATLONG, datum=WGS84)
    lon, lat = transform(p1, p2, east, north)
    return lat, lon


def anti_alias(input):
    output = input.copy()
    for dx in (-0.5, 0.5):
        for dy in (-0.5, 0.5):
            mask = shift(input, (dx, dy), np.float)
            output = np.maximum(output, mask)
    return output


def overlay(image, mask, rgb):
    inverse = 1 - mask
    data = extract_rgb(image)
    data = [(np.clip(layer * inverse + mask * color * np.max(layer), 0, np.max(layer))).astype(layer.dtype)
            for layer, color in zip(data, rgb)]
    return create_image(image, np.stack(data))


def overlay_route(image, lat, lon, rgb):
    x, y = latlon_to_xy(image, lat, lon)
    mask = np.zeros((image.height, image.width), dtype=np.float)
    rows, columns = polygon_perimeter(y, x, shape=mask.shape, clip=False)
    mask[rows, columns] = 1
    mask = anti_alias(mask)
    return overlay(image, mask, rgb)






import datetime as dt
from collections import namedtuple
from json import loads
from logging import getLogger

import numpy as np
from math import log10
from sqlalchemy.sql.functions import count

from . import UniProcCalculator
from ..read.segment import SegmentReader
from ...data.frame import statistics
from ...data.response import sum_to_hour, calc_response
from ...lib.date import round_hour, to_time, local_date_to_time, now
from ...sql import ActivityGroup
from ...sql import StatisticJournal, Composite, StatisticName, Source, Constant, CompositeComponent, \
    StatisticJournalFloat
from ...sql.utils import add
from ...stats.names import _src, ALL, HR_IMPULSE_10, COVERAGE, HEART_RATE
from ...stats.pipeline import LoaderMixin

log = getLogger(__name__)

Response = namedtuple('Response', 'src_owner, dest_name, tau_days, start, scale')

SCALED = 'Scaled'


class ResponseCalculator(LoaderMixin, UniProcCalculator):
    '''
    this is hard to do correctly, incrementally.

    for now, we do all or nothing.

    first, we check if the current solution is complete:
    * all sources used (no need to check for gaps - composite chaining should do that)
    * within 3 hours of the current time (if not, we regenerate, padding with zeroes)

    if not complete, we regenerate the whole damn thing.
    '''

    def __init__(self, *args, responses_ref=None, **kargs):
        self.responses_ref = self._assert('responses_ref', responses_ref)
        super().__init__(*args, **kargs)

    def _startup(self, s):
        super()._startup(s)
        constants = [Constant.get(s, response) for response in self.responses_ref]
        self.responses = [Response(**loads(constant.at(s).value)) for constant in constants]

    def _delete(self, s):
        start, finish = self._start_finish(local_date_to_time)
        if start or self.force:
            if finish:
                log.warning(f'Ignoring finish - will delete all impulses from {start} onwards')
            self._delete_from(s, start)

    def _delete_from(self, s, start=None, inclusive=True):
        composite_ids = s.query(Composite.id). \
            join(StatisticJournal, Composite.id == StatisticJournal.source_id). \
            join(StatisticName, StatisticJournal.statistic_name_id == StatisticName.id). \
            filter(StatisticName.owner == self.owner_out)
        if start:
            if inclusive:
                composite_ids = composite_ids.filter(StatisticJournal.time >= start)
            else:
                composite_ids = composite_ids.filter(StatisticJournal.time > start)
        log.debug(f'Delete query: {composite_ids}')
        n = s.query(count(Source.id)). \
            filter(Source.id.in_(composite_ids)). \
            scalar()
        if n:
            log.warning(f'Deleting {n} Composite sources ({start} onwards{" inclusive" if inclusive else ""})')
            s.query(Source). \
                filter(Source.id.in_(composite_ids)). \
                delete(synchronize_session=False)
            s.commit()

    def _missing(self, s):
        # clean out any gaps by unzipping the chained composites
        Composite.clean(s)
        # range we expect data for
        missing_sources = self.__missing_sources(s)
        finish = round_hour(dt.datetime.now(tz=dt.timezone.utc), up=True)
        missing_recent = any(self.__missing_recent(s, response.dest_name, finish) for response in self.responses)
        if missing_recent or missing_sources:
            if missing_recent: log.info('Incomplete coverage (so will re-calculate)')
            if missing_sources: log.info('Additional sources (so will re-calculate)')
            self._delete_from(s)
            start = round_hour(self.__start(s), up=False)
            return [(start, finish)]
        else:
            return []

    def __missing_recent(self, s, response, now):
        finish = s.query(StatisticJournal.time). \
            join(StatisticName, StatisticJournal.statistic_name_id == StatisticName.id). \
            filter(StatisticName.name == response,
                   StatisticName.owner == self.owner_out). \
            order_by(StatisticJournal.time.desc()).limit(1).one_or_none()
        if finish is None:
            log.debug(f'Missing data for {response}')
            return True
        if (now - finish[0]).total_seconds() > 3 * 60 * 60:
            log.debug(f'Insufficient data for {response}')
            return True
        return False

    def __missing_sources(self, s):
        response_ids = s.query(StatisticName.id). \
            filter(StatisticName.owner == self.owner_out)
        inputs = s.query(CompositeComponent.input_source_id). \
            join(StatisticJournal, StatisticJournal.source_id == CompositeComponent.output_source_id). \
            filter(StatisticJournal.statistic_name_id.in_(response_ids),
                   CompositeComponent.input_source_id != None)
        unused = s.query(count(StatisticJournal.id)). \
            join(StatisticName, StatisticJournal.statistic_name_id == StatisticName.id). \
            filter(StatisticName.name == HR_IMPULSE_10,
                   StatisticName.owner == self.owner_in,
                   ~StatisticJournal.source_id.in_(inputs))
        log.debug(unused)
        log.debug(f'owner_in: {self.owner_in}')
        log.debug(f'owner_out: {self.owner_out}')
        return unused.scalar()

    def __start(self, s):
        # avoid constants defined at time 0
        return s.query(StatisticJournal.time). \
            filter(StatisticJournal.time > 1.0). \
            order_by(StatisticJournal.time.asc()). \
            limit(1).scalar()

    def _run_one(self, s, missed):
        data = self.__read_data(s)
        if HR_IMPULSE_10 in data.columns and COVERAGE in data.columns:
            data.loc[now()] = {HR_IMPULSE_10: 0.0, _src(HR_IMPULSE_10): None, COVERAGE: 100}
            data[SCALED] = data[HR_IMPULSE_10] * 100 / data[COVERAGE]
            all_sources = list(self.__make_sources(s, data))
            for response in self.responses:
                log.info(f'Creating values for {response.dest_name}')
                hr3600 = sum_to_hour(data, SCALED)
                params = (log10(response.tau_days * 24),
                          log10(response.start) if response.start > 0 else 1)
                result = calc_response(hr3600, params) * response.scale
                loader = self._get_loader(s, add_serial=False)
                source, sources = None, list(all_sources)
                for time, value in result.iteritems():
                    while sources and time >= sources[0][0]:
                        source = sources.pop(0)[1]
                    loader.add(response.dest_name, None, None, None, source, value, time, StatisticJournalFloat)
                loader.load()

    def __read_coverage(self, s):
        names = s.query(StatisticName).\
            filter(StatisticName.name == COVERAGE).\
            filter(StatisticName.constraint.like(HEART_RATE + ' / %')).all()
        coverages = statistics(s, *names, owner=SegmentReader, check=False)
        coverages = coverages.loc[:].replace(0, np.nan)
        coverages.fillna(axis='columns', method='bfill', inplace=True)
        coverages.fillna(axis='columns', method='ffill', inplace=True)
        if not coverages.empty:
            coverages = coverages.iloc[:, [0]].rename(columns={coverages.columns[0]: COVERAGE})
        return coverages

    def __read_data(self, s):
        hr10 = statistics(s, HR_IMPULSE_10, constraint=ActivityGroup.from_name(s, ALL),
                          owner=self.owner_in, with_sources=True, check=False)
        coverage = self.__read_coverage(s)
        coverage.reindex(index=hr10.index, method='nearest', copy=False)
        data = hr10.join(coverage, how='outer')
        if HR_IMPULSE_10 in data.columns and COVERAGE in data.columns:
            data.loc[:, [HR_IMPULSE_10]] = data.loc[:, [HR_IMPULSE_10]].fillna(0.0,)
            data.loc[:, [COVERAGE]] = data.loc[:, [COVERAGE]].fillna(axis='index', method='ffill')
            data.loc[:, [COVERAGE]] = data.loc[:, [COVERAGE]].fillna(axis='index', method='bfill')
        return data

    def __make_sources(self, s, data):
        log.info('Creating sources')
        name = _src(HR_IMPULSE_10)
        prev = add(s, Composite(n_components=0))
        yield to_time(0.0), prev
        # find times where the source changes
        for time, row in data.loc[data[name].ne(data[name].shift())].iterrows():
            id = row[name]
            if id:
                composite = add(s, Composite(n_components=2))
                add(s, CompositeComponent(input_source_id=id, output_source=composite))
            else:
                composite = add(s, Composite(n_components=1))
            add(s, CompositeComponent(input_source=prev, output_source=composite))
            yield time, composite
            prev = composite
        s.commit()




import re
from logging import getLogger

'''
The model is intended to be, as far as reasonable, self-describing.
In other words, clients are expected to be able to display the model even if the order of the data changes,
or if entries are added or removed.
Exceptions to the above should be keyed on tag attributes (see below).
The 'low-level' details (eg attribute names) of the model should not change.

The model is hierarchical, modelled as a tree.
So the outermost list is the 'root' and each entry is a 'node'.
Non-leaf nodes, including  the root, are lists.
Leaf nodes are dicts / maps and represent individual 'values'.

Within a non-leaf node, the first child node is a 'text' type node (see below) that acts as a 'title' 
describing the contents of the node.

All leaf nodes have a 'type' attribute which describes the type of the node.
The contents of a node of a given type match the constructor functions below.
All leaf nodes have a 'value' attribute.
All leaf nodes except those of type 'text' have a 'label' attribute.
All leaf nodes have a 'tag' attribute that is intended as a machine-readable semantic marker (so where the client 
needs exceptional processing I will try to maintain these even if the text of a node changes, for example).
'''

# diary and general
DB = 'db'
DP = 'dp'
EDIT = 'edit'
FLOAT = 'float'
HI = 'hi'
INTEGER = 'integer'
LABEL = 'label'
LINK = 'link'
LO = 'lo'
MEASURES = 'measures'
TAG = 'tag'
SCHEDULES = 'schedules'
SCORE = 'score'
TEXT = 'text'
TYPE = 'type'
UNITS = 'units'
VALUE = 'value'


log = getLogger(__name__)


def from_field(topic_field, statistic_journal):
    kargs = dict(topic_field.model)
    type = kargs[TYPE]
    del kargs[TYPE]
    kargs.update(value=statistic_journal.value,
                 label=statistic_journal.statistic_name.name,
                 db=statistic_journal)
    if statistic_journal.statistic_name.units:
        kargs.update(units=statistic_journal.statistic_name.units)
    return {SCORE: score, INTEGER: integer, FLOAT: float, EDIT: edit}[type](**kargs)


def to_tag(text):
    text = re.sub(r'\W+', '-', text)
    text = re.sub(r'-?(\w+(:?-\w+)*)-?', r'\1', text)
    return text.lower()


# --- mutable types

def score(label, value, db=None):
    if db is None: log.warning(f'No db for score {label}/{value}')
    return {TYPE: SCORE, LABEL: label, VALUE: value, DB: db}


def integer(label, value, units=None, lo=None, hi=None, db=None):
    if db is None: log.warning(f'No db for integer {label}/{value}')
    return {TYPE: INTEGER, LABEL: label, VALUE: value, UNITS: units, LO: lo, HI: hi, DB: db}


def float(label, value, units=None, lo=None, hi=None, dp=1, db=None):
    if db is None: log.warning(f'No db for float {label}/{value}')
    return {TYPE: FLOAT, LABEL: label, VALUE: value, UNITS: units, LO: lo, HI: hi, DP: dp, DB: db}


def edit(label, value, db=None):
    if db is None: log.warning(f'No db for edit {label}/{value}')
    return {TYPE: EDIT, LABEL: label, VALUE: value, DB: db}


# --- immutable types

def text(value: object, tag: object = None) -> object:
    return {TYPE: TEXT, VALUE: value, TAG: to_tag(tag or value)}


def value(label, value, tag=None, units=None, measures=None):
    return {TYPE: VALUE, LABEL: label, VALUE: value, TAG: to_tag(tag or label), UNITS: units, MEASURES: measures}


def measures(schedules):
    # schedules are a map from schedule to (percent, rank) tuples
    return {TYPE: MEASURES, SCHEDULES: schedules}


def link(value, tag=None, db=None):
    if not isinstance(db, tuple): log.warning(f'Bad db for link {value}')
    return {TYPE: LINK, VALUE: value, TAG: to_tag(tag or value), DB: db}


# --- decorators

def optional_text(name, tag=None):
    def decorator(f):
        def decorated(*args, **kargs):
            first = True
            for value in f(*args, **kargs):
                if first:
                    yield text(name, tag=tag)
                    first = False
                yield value
        return decorated
    return decorator


def trim_no_stats(f):

    def decorated(*args, **kargs):
        result = list(f(*args, **kargs))

        def trim(model):
            if isinstance(model, list):
                head, rest = model[0:1], model[1:]
                rest = [x for x in [trim(entry) for entry in rest] if x]
                if rest:
                    return head + rest
                else:
                    return []
            else:
                return model
        return trim(result)

    return decorated




import datetime as dt
from enum import IntEnum
from logging import getLogger
from re import split

from sqlalchemy import Column, Integer, ForeignKey, Text, UniqueConstraint, Float, desc, asc, Index
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm import relationship, backref, reconstructor
from sqlalchemy.orm.exc import NoResultFound

from .source import Interval
from ..support import Base
from ..types import Time, ShortCls, NullStr
from ..utils import add
from ...diary.model import TYPE, MEASURES, SCHEDULES
from ...lib.date import format_seconds, local_date_to_time
from ...lib.utils import sigfig
from ...stats.names import KMH, PC, BPM, STEPS_UNITS, S, M, KG, W, KCAL, KJ, FF

log = getLogger(__name__)


class StatisticName(Base):

    __tablename__ = 'statistic_name'

    id = Column(Integer, primary_key=True)
    name = Column(Text, nullable=False, index=True)  # simple, displayable name
    description = Column(Text)
    units = Column(Text)
    summary = Column(Text)  # '[max]', '[min]' etc - can be multiple values but each in square brackets
    # we need to disambiguate statistics with the same name.
    # this is done by (1) "owner" (typically the source of the data) and
    # (2) by some additional (optional) constraint used by the owner (typically)
    # (eg activity_group.id so that the same statistic can be used across different activities)
    owner = Column(ShortCls, nullable=False, index=True)  # index for deletion
    constraint = Column(NullStr)
    statistic_journal_type = Column(Integer, nullable=False)  # StatisticJournalType
    UniqueConstraint(name, owner, constraint)

    def __str__(self):
        return '"%s" (%s/%s)' % (self.name, self.owner, self.constraint)

    @property
    def summaries(self):
        if self.summary:
            return [x.lower() for x in split(r'[\s,]*(\[[^\]]+\])[\s ]*', self.summary) if x]
        else:
            return []

    @classmethod
    def add_if_missing(cls, s, name, type, units, summary, owner, constraint, description=None):
        s.commit()  # start new transaction here in case rollback
        q = s.query(StatisticName). \
            filter(StatisticName.name == name,
                   StatisticName.owner == owner,
                   StatisticName.constraint == constraint)
        statistic_name = q.one_or_none()
        if not statistic_name:
            statistic_name = add(s, StatisticName(name=name, units=units, summary=summary, owner=owner,
                                                  constraint=constraint, statistic_journal_type=type,
                                                  description=description))
            try:
                s.flush()
            except IntegrityError as e:  # worker may have created in parallel, so read
                log.debug(f'Rollback for {e}')
                s.rollback()
                log.debug('Now trying retrieval...')
                statistic_name = q.one()
                log.debug('Retrieved')
        else:
            if statistic_name.statistic_journal_type != type:
                raise Exception('Changing type on %s (%s -> %s)' %
                                (statistic_name.name, statistic_name.statistic_journal_type, type))
            if statistic_name.units != units:
                log.warning('Changing units on %s (%s -> %s)' % (statistic_name.name, statistic_name.units, units))
                statistic_name.units = units
                s.flush()
            if statistic_name.summary != summary:
                log.warning('Changing summary on %s (%s -> %s)' % (statistic_name.name, statistic_name.summary, summary))
                statistic_name.summary = summary
                s.flush()
            if statistic_name.description != description:
                log.warning('Changing description on %s (%s -> %s)' % (statistic_name.name, statistic_name.description, description))
                statistic_name.description = description
                s.flush()
        return statistic_name

    @classmethod
    def from_name(cls, s, name, owner, constaint=None):
        return s.query(StatisticName). \
            filter(StatisticName.name == name,
                   StatisticName.owner == owner,
                   StatisticName.constraint == constaint).one()

    @classmethod
    def parse(cls, name, default_owner=None, default_constraint=None):
        '''
        This parses the standard, extended format for naming statistics.  It is one to three fields, separated by ':'.
        These are one of 'name', 'owner:name', or 'owner:name:constraint'.
        Currently this is used only by reftuple which itself is used only in the power pipeline configuration.
        '''
        parts, owner, constraint = name.split(':'), None, None
        if len(parts) == 1:
            name = parts[0]
        elif len(parts) == 2:
            owner, name = parts
        else:
            owner, name, constraint = parts
        if not owner: owner = default_owner
        if not owner:
            raise Exception(f'Missing owner for {name}')
        if not constraint: constraint = default_constraint
        if constraint == 'None': constraint = None
        return owner, name, constraint


class StatisticJournalType(IntEnum):

    STATISTIC = 0
    INTEGER = 1
    FLOAT = 2
    TEXT = 3
    TIMESTAMP = 4


class StatisticJournal(Base):

    __tablename__ = 'statistic_journal'

    id = Column(Integer, primary_key=True)
    type = Column(Integer, nullable=False, index=True)  # index needed for fast delete of subtypes
    statistic_name_id = Column(Integer, ForeignKey('statistic_name.id', ondelete='cascade'),
                               nullable=False, index=True)
    statistic_name = relationship('StatisticName')
    source_id = Column(Integer, ForeignKey('source.id', ondelete='cascade'),
                       nullable=False, index=True)
    source = relationship('Source')
    time = Column(Time, nullable=False)
    # serial "counts" along values in the timeseries.  it's optional.  for garmin, all values appear each
    # record, so all imported values share the same serial.  but that's not true for the corrected elevation,
    # for example.
    serial = Column(Integer)
    # relax this for kit support (add source_id)
    UniqueConstraint(statistic_name_id, time, source_id)
    UniqueConstraint(serial, source_id, statistic_name_id)
    Index('from_activity_timespan', source_id, statistic_name_id, time)  # time last since inequality

    __mapper_args__ = {
        'polymorphic_identity': StatisticJournalType.STATISTIC,
        'polymorphic_on': 'type'
    }

    def __str__(self):
        try:
            return 'StatisticJournal "%s"' % self.value
        except AttributeError:
            return 'StatisticJournal base'

    @classmethod
    def add(cls, s, name, units, summary, owner, constraint, source, value, time, serial, type):
        statistic_name = StatisticName.add_if_missing(s, name, type, units, summary, owner, constraint)
        journal = STATISTIC_JOURNAL_CLASSES[type](
                statistic_name=statistic_name, source=source, value=value, time=time, serial=serial)
        s.add(journal)
        return journal

    def formatted(self):
        if self.value is None:
            return None
        units = self.statistic_name.units
        if not units:
            return '%d' % self.value
        elif units == M:
            if self.value > 2000:
                return '%d km' % (self.value / 1000)
            else:
                return '%d m' % self.value
        elif units == S:
            return format_seconds(self.value)
        elif units in (KMH, PC, BPM, STEPS_UNITS, W, KJ):
            return '%d %s' % (self.value, units)
        elif units == KCAL:
            return '%s %s' % (sigfig(self.value, 2), units)
        else:
            return '%d %s' % (self.value, units)

    def measures_as_text(self, date):
        words = []
        if hasattr(self, 'measures'):
            for measure in sorted(self.measures,
                                  key=lambda measure: measure.source.schedule.frame_length_in_days(date)):
                if words:
                    words += [',']
                quintile = 1 + min(4, measure.percentile / 20)
                words += [('quintile-%d' % quintile, '%d%%' % int(measure.percentile))]
                if measure.rank < 5:
                    words += [':', ('rank-%d' % measure.rank, '%d' % measure.rank)]
                words += ['/' + measure.source.schedule.describe(compact=True)]
        return words

    def measures_as_model(self, date):
        if hasattr(self, 'measures'):
            measures = {TYPE: MEASURES, SCHEDULES: {}}
            for measure in sorted(self.measures,
                                  key=lambda measure: measure.source.schedule.frame_length_in_days(date),
                                  reverse=True):
                measures[SCHEDULES][measure.source.schedule.describe(compact=True)] = (measure.percentile, measure.rank)
            return measures
        else:
            return None

    @classmethod
    def at_date(cls, s, date, name, owner, constraint, source_id=None):
        start = local_date_to_time(date)
        finish = start + dt.timedelta(days=1)
        q = s.query(StatisticJournal).join(StatisticName). \
            filter(StatisticName.name == name,
                   StatisticJournal.time >= start,
                   StatisticJournal.time < finish,
                   StatisticName.owner == owner,
                   StatisticName.constraint == constraint)
        if source_id is not None:
            q = q.filter(StatisticJournal.source_id == source_id)
        return q.one_or_none()

    @classmethod
    def at(cls, s, time, name, owner, constraint):
        return s.query(StatisticJournal).join(StatisticName). \
            filter(StatisticName.name == name,
                   StatisticJournal.time == time,
                   StatisticName.owner == owner,
                   StatisticName.constraint == constraint).one_or_none()

    @classmethod
    def at_like(cls, s, time, name, owner, constraint):
        return s.query(StatisticJournal).join(StatisticName). \
            filter(StatisticName.name.like(name),
                   StatisticJournal.time == time,
                   StatisticName.owner == owner,
                   StatisticName.constraint == constraint).all()

    @classmethod
    def before(cls, s, time, name, owner, constraint):
        return s.query(StatisticJournal).join(StatisticName). \
            filter(StatisticName.name == name,
                   StatisticJournal.time <= time,
                   StatisticName.owner == owner,
                   StatisticName.constraint == constraint). \
            order_by(desc(StatisticJournal.time)).limit(1).one_or_none()

    @classmethod
    def before_not_null(cls, s, time, name, owner, constraint):
        try:
            statistic_name = s.query(StatisticName). \
                filter(StatisticName.name == name,
                       StatisticName.owner == owner,
                       StatisticName.constraint == constraint).one()
        except NoResultFound:
            raise Exception(f'The statistic name "{name}" (owner {owner}, constraint {constraint}) '
                            'is undefined in the database')
        cls = STATISTIC_JOURNAL_CLASSES[statistic_name.statistic_journal_type]
        return s.query(cls). \
            filter(cls.statistic_name == statistic_name,
                   cls.time <= time,
                   cls.value != None). \
            order_by(desc(cls.time)).limit(1).one_or_none()

    @classmethod
    def after(cls, s, time, name, owner, constraint):
        return s.query(StatisticJournal).join(StatisticName). \
            filter(StatisticName.name == name,
                   StatisticJournal.time >= time,
                   StatisticName.owner == owner,
                   StatisticName.constraint == constraint). \
            order_by(asc(StatisticJournal.time)).limit(1).one_or_none()

    @classmethod
    def at_interval(cls, s, start, schedule, statistic_owner, statistic_constraint, interval_owner):
        return s.query(StatisticJournal).join(StatisticName, Interval). \
                    filter(StatisticJournal.statistic_name_id == StatisticName.id,
                           Interval.schedule == schedule,
                           Interval.start == start,
                           Interval.owner == interval_owner,
                           StatisticName.owner == statistic_owner,
                           StatisticName.constraint == statistic_constraint). \
                    order_by(StatisticName.constraint,  # order places summary stats from same source together
                             StatisticName.name).all()

    @classmethod
    def for_source(cls, s, source_id, name, owner, constraint):
        return s.query(StatisticJournal).join(StatisticName). \
            filter(StatisticName.name == name,
                   StatisticName.owner == owner,
                   StatisticName.constraint == constraint,
                   StatisticJournal.source_id == source_id).one_or_none()


class StatisticJournalInteger(StatisticJournal):

    __tablename__ = 'statistic_journal_integer'

    id = Column(Integer, ForeignKey('statistic_journal.id', ondelete='cascade'), primary_key=True)
    value = Column(Integer)

    __mapper_args__ = {
        'polymorphic_identity': StatisticJournalType.INTEGER
    }

    def set(self, value):
        self.value = value if value is None else int(value)

    @classmethod
    def add(cls, s, name, units, summary, owner, constraint, source, value, time, serial=None):
        return super().add(s, name, units, summary, owner, constraint, source, value, time, serial,
                           StatisticJournalType.INTEGER)


class StatisticJournalFloat(StatisticJournal):

    __tablename__ = 'statistic_journal_float'

    id = Column(Integer, ForeignKey('statistic_journal.id', ondelete='cascade'), primary_key=True)
    value = Column(Float)

    def set(self, value):
        self.value = value if value is None else float(value)

    @classmethod
    def add(cls, s, name, units, summary, owner, constraint, source, value, time, serial=None):
        return super().add(s, name, units, summary, owner, constraint, source, value, time, serial,
                           StatisticJournalType.FLOAT)

    __mapper_args__ = {
        'polymorphic_identity': StatisticJournalType.FLOAT
    }

    def formatted(self):
        if self.value is None:
            return None
        units = self.statistic_name.units
        if not units:
            return '%f' % self.value
        elif units == M:
            if self.value > 2000:
                return '%.1f km' % (self.value / 1000)
            else:
                return '%d m' % int(self.value)
        elif units == S:
            return format_seconds(self.value)
        elif units in (KMH, PC, KG, W, KJ):
            return '%.1f %s' % (self.value, units)
        elif units == KCAL:
            return '%s %s' % (sigfig(self.value, 2), units)
        elif units in (BPM, STEPS_UNITS):
            return '%d %s' % (int(self.value), units)
        elif units == FF:
            return '%d' % int(self.value)
        else:
            return '%s %s' % (self.value, units)


class StatisticJournalText(StatisticJournal):

    __tablename__ = 'statistic_journal_text'

    id = Column(Integer, ForeignKey('statistic_journal.id', ondelete='cascade'), primary_key=True)
    value = Column(Text)

    def set(self, value):
        self.value = value if value is None else str(value)

    @classmethod
    def add(cls, s, name, units, summary, owner, constraint, source, value, time, serial=None):
        return super().add(s, name, units, summary, owner, constraint, source, value, time, serial,
                           StatisticJournalType.TEXT)

    __mapper_args__ = {
        'polymorphic_identity': StatisticJournalType.TEXT
    }

    def formatted(self):
        if self.value is None:
            return None
        if not self.units:
            return '%s' % self.value
        else:
            return '%s %s' % (self.value, self.units)


class StatisticJournalTimestamp(StatisticJournal):

    # todo - could value take value from time?

    __tablename__ = 'statistic_journal_timestamp'

    id = Column(Integer, ForeignKey('statistic_journal.id', ondelete='cascade'), primary_key=True)

    __mapper_args__ = {
        'polymorphic_identity': StatisticJournalType.TIMESTAMP
    }

    def __init__(self, **kargs):
        super().__init__(**kargs)
        self.value = None

    @reconstructor
    def init_on_load(self):
        # we need a value attribute so that the API is constant across all journal instances
        self.value = None

    @classmethod
    def add(cls, s, name, units, summary, owner, constraint, source, time, serial=None):
        statistic_name = StatisticName.add_if_missing(s, name, StatisticJournalType.TIMESTAMP,
                                                      units, summary, owner, constraint)
        journal = StatisticJournalTimestamp(statistic_name=statistic_name, source=source, time=time, serial=serial)
        s.add(journal)
        return journal


class StatisticMeasure(Base):

    __tablename__ = 'statistic_measure'

    id = Column(Integer, primary_key=True)
    statistic_journal_id = Column(Integer, ForeignKey('statistic_journal.id', ondelete='cascade'),
                                  nullable=False, index=True)
    statistic_journal = relationship('StatisticJournal',
                                     backref=backref('measures', cascade='all, delete-orphan',
                                                     passive_deletes=True,
                                                     order_by='desc(StatisticMeasure.rank)'))
    source_id = Column(Integer, ForeignKey('source.id', ondelete='cascade'),
                       nullable=False, index=True)  # must be an interval
    source = relationship('Source')
    rank = Column(Integer, nullable=False)  # 1 is best [1..n]
    percentile = Column(Float, nullable=False)  # 100 is best [0..100]
    quartile = Column(Integer)  # 0..4 at the min, 25%, median, 75% and max points


STATISTIC_JOURNAL_CLASSES = {
    StatisticJournalType.INTEGER: StatisticJournalInteger,
    StatisticJournalType.FLOAT: StatisticJournalFloat,
    StatisticJournalType.TEXT: StatisticJournalText
}

STATISTIC_JOURNAL_TYPES = {
    StatisticJournalInteger: StatisticJournalType.INTEGER,
    StatisticJournalFloat: StatisticJournalType.FLOAT,
    StatisticJournalText: StatisticJournalType.TEXT
}

TYPE_TO_JOURNAL_CLASS = {
    int: StatisticJournalInteger,
    float: StatisticJournalFloat,
    str: StatisticJournalText
}




from json import loads
from tempfile import TemporaryDirectory
from tests import LogTestCase

from ch2.commands.args import bootstrap_dir, m, V
from ch2.lib.data import MutableAttr, reftuple
from ch2.sql import StatisticJournalFloat, StatisticJournalText, Source
from ch2.sql.tables.source import SourceType


class TestData(LogTestCase):

    def test_attr(self):
        d = MutableAttr()
        d['foo'] = 'bar'
        self.assertTrue('foo' in d)
        self.assertEqual(d.foo, 'bar')

    def test_reftuple(self):

        Power = reftuple('Power', 'bike, weight')
        power = Power('${Bike}', '${Weight}')

        with TemporaryDirectory() as f:
            args, sys, db = bootstrap_dir(f, m(V), '5')
            with db.session_context() as s:
                source = Source(type=SourceType.SOURCE)
                s.add(source)
                StatisticJournalText.add(s, 'Bike', None, None, self, None, source, '{"mass": 42}', '1980-01-01')
                StatisticJournalFloat.add(s, 'Weight', None, None, self, None, source, 13, '1980-01-01')
            p = power.expand(s, '1990-01-01', default_owner=self)
            p = p._replace(bike=loads(p.bike))
            self.assertEqual(p.weight, 13)
            self.assertEqual(p.bike['mass'], 42)

    def test_types(self):
        from ch2.stats.calculate.power import Power, Bike
        self.assertEqual(Bike.__module__, 'ch2.stats.calculate.power')
        self.assertEqual(Power.__module__, 'ch2.stats.calculate.power')




from abc import abstractmethod, ABC
from logging import getLogger
from sys import stdout
import re

from .args import TOPIC, HELP, PROGNAME, m, H, VERSION
from ..lib.io import terminal_width

log = getLogger(__name__)


def commands(COMMANDS):
    return '''
Thank-you for using Choochoo.  Please send feedback to andrew@acooke.org

# Commands

* %s

See also `%s %s` for usage, '%s %s CMD` for guidance on a particular command, 
and `%s %s CMD` for usage of that command.

Docs at http://andrewcooke.github.io/choochoo/index''' % (
        '\n* '.join(COMMANDS.keys()), PROGNAME, m(H), PROGNAME, HELP, PROGNAME, m(H))


def help(args, system, db):
    '''
## help

    > ch2 help [topic]

Displays help for a particular topic.

### Examples

    > ch2 help help

Displays this information.

    > ch2 help

Lists available topics.
    '''
    from .. import COMMANDS
    if args[TOPIC] in COMMANDS:
        Markdown().print(COMMANDS[args[TOPIC]].__doc__)
    else:
        Markdown().print(commands(COMMANDS))


BR = 'BR'
H = 'H'
P = 'P'
LI = 'LI'
PRE = 'PRE'


def parse(text):
    '''
    A very simple parser for the tiny subset of markdown that is used in comments.
    '''
    paragraph = ''
    for line in text.split('\n'):
        if line.strip().startswith('*'):
            if paragraph:
                yield P, paragraph
                paragraph = ''
            yield LI, line.strip().split(' ', 1)[1]
        elif line.strip() and line.startswith('  '):
            if paragraph:
                yield P, paragraph
                paragraph = ''
            yield PRE, line.strip()
        elif line.startswith('#'):
            if paragraph:
                yield P, paragraph
                paragraph = ''
            h, line = line.split(' ', 1)
            yield H + str(len(h)), line.strip()
        elif not line.strip():
            if paragraph:
                yield P, paragraph
                paragraph = ''
            yield BR, None
        elif paragraph:
            paragraph += ' ' + line.strip()
        else:
            paragraph = line
    if paragraph:
        yield P, paragraph


def filter(parser, yes=None, no=None):

    def filtered(text):
        for tag, value in parser(text):
            if yes and tag in yes:
                yield tag, value
            elif no and tag not in no:
                yield tag, value
            else:
                pass

    return filtered


class Formatter(ABC):

    def __init__(self, parser=None):
        self._parser = parser if parser else parse

    @abstractmethod
    def p(self, text):
        raise NotImplementedError

    @abstractmethod
    def li(self, text):
        raise NotImplementedError

    @abstractmethod
    def pre(self, text):
        raise NotImplementedError

    @abstractmethod
    def h(self, level, text):
        raise NotImplementedError

    @abstractmethod
    def br(self):
        raise NotImplementedError

    def format(self, text):
        for tag, text in self._parser(text):
            if tag == P: yield from self.p(text)
            elif tag == LI: yield from self.li(text)
            elif tag == PRE: yield from self.pre(text)
            elif tag == BR: yield from self.br()
            else: yield from self.h(int(tag[1:]), text)

    def str(self, text):
        return '\n'.join(self.format(text))

    def print(self, text, out=stdout):
        print(self.str(text), file=out)


class Markdown(Formatter):

    def __init__(self, width=None, margin=1, delta=0, parser=None):
        super().__init__(parser=parser)
        self._width = terminal_width(width) - margin
        self._margin = margin
        self._delta = delta

    def p(self, text):
        line = ''
        for word in re.sub(r' +', ' ', text).split(' '):
            if not line: line = ' ' * self._margin
            else: line += ' '
            if len(line + word) > self._width:
                yield line
                line = ' ' * self._margin + word
            else:
                line += word
        if line: yield line

    def li(self, text):
        yield (' ' * self._margin) + '* ' + text

    def pre(self, text):
        yield (' ' * self._margin) + '    ' + text

    def h(self, level, text):
        yield (' ' * self._margin) + '#' * max(1, level + self._delta) + ' ' + text

    def br(self):
        yield ''


class HTML(Formatter):

    def __init__(self, delta=0, parser=None):
        super().__init__(parser=parser)
        self._delta = delta
        self._list = False

    def _ul(self, list):
        if self._list != list:
            if list: yield '<ul>'
            else: yield '</ul>'
            self._list = list

    def _tag(self, tag, text):
        yield f'<{tag}>'
        yield text
        yield f'</{tag}>'

    def p(self, text):
        yield from self._ul(False)
        yield from self._tag('p', text)

    def li(self, text):
        yield from self._ul(True)
        yield from self._tag('li', text)

    def pre(self, text):
        yield from self._ul(False)
        yield from self._tag('pre', text)

    def h(self, level, text):
        yield from self._ul(False)
        yield from self._tag('h' + str(max(1, level + self._delta)), text)

    def br(self):
        return
        yield



from io import StringIO
from tempfile import TemporaryDirectory

from ch2.commands.activities import activities
from ch2.commands.args import bootstrap_dir, m, V, mm, DEV, D, BASE
from ch2.commands.kit import start, change, finish, show, undo, statistics
from ch2.config import default
from ch2.diary.model import TYPE
from ch2.lib import now, local_date_to_time
from ch2.sql import KitModel, KitItem, KitComponent, PipelineType
from ch2.sql.tables.kit import get_name, KitGroup, NAME, ITEMS, COMPONENTS, MODELS, STATISTICS, INDIVIDUAL
from ch2.stats.pipeline import run_pipeline
from tests import LogTestCase


def days(date):
    return (now() - local_date_to_time(date)).days


class TestKit(LogTestCase):

    def test_bikes(self):
        with TemporaryDirectory() as f:
            args, sys, db = bootstrap_dir(f, m(V), '5', configurator=default)
            with db.session_context() as s:
                with self.assertRaises(Exception) as ctx:
                    start(s, 'bike', 'cotic', '2020-03-24', False)
                self.assertTrue('--force' in str(ctx.exception), ctx.exception)
                start(s, 'bike', 'cotic', '2020-03-24', True)
                # run('sqlite3 %s ".dump"' % f.name, shell=True)
                with self.assertRaises(Exception) as ctx:
                    start(s, 'xxxx', 'marin', None, False)
                self.assertTrue('--force' in str(ctx.exception), ctx.exception)
                start(s, 'bike', 'marin', None, False)
                with self.assertRaises(Exception) as ctx:
                    change(s, 'cotic', 'chain', 'sram', '2018-02-01', False, False)
                self.assertTrue('--force' in str(ctx.exception))
                change(s, 'cotic', 'chain', 'sram', None, True, True)
                change(s, 'cotic', 'chain', 'kcm', '2018-01-01', False, False)
                change(s, 'cotic', 'chain', 'sram', '2018-05-01', False, False)
                change(s, 'cotic', 'chain', 'kcm', '2018-07-01', False, False)
                change(s, 'cotic', 'chain', 'sram', '2018-04-01', False, False)
                with self.assertRaises(Exception) as ctx:
                    start(s, 'bike', 'bike', None, True)
                self.assertTrue('bike' in str(ctx.exception), ctx.exception)
                with self.assertRaises(Exception) as ctx:
                    start(s, 'bike', 'sram', None, True)
                self.assertTrue('sram' in str(ctx.exception), ctx.exception)
                start(s, 'bike', 'bowman', None, False)
                change(s, 'bowman', 'chain', 'sram', None, False, True)
                self.assert_command('''item: cotic  2020-03-24 - 
`-component: chain
  +-model: sram  2020-03-24 - 
  +-model: kcm  2018-01-01 - 2018-04-01
  +-model: sram  2018-05-01 - 2018-07-01
  +-model: kcm  2018-07-01 - 2020-03-24
  `-model: sram  2018-04-01 - 2018-05-01
''', show, s, 'cotic', None)
                self.assert_command(f'''group: bike
+-item: cotic
| +-Age
| | +-n: 1
| | `-sum: {days('2020-03-24')}
| `-component: chain
|   +-model: sram
|   | `-Age
|   |   +-n: 1
|   |   `-sum: {days('2020-03-24')}
|   +-model: kcm
|   | `-Age
|   |   +-n: 1
|   |   `-sum: 90
|   +-model: sram
|   | `-Age
|   |   +-n: 1
|   |   `-sum: 61
|   +-model: kcm
|   | `-Age
|   |   +-n: 1
|   |   `-sum: 631
|   `-model: sram
|     `-Age
|       +-n: 1
|       `-sum: 30
+-item: marin
| `-Age
|   +-n: 1
|   `-sum: 0
`-item: bowman
  +-Age
  | +-n: 1
  | `-sum: 0
  `-component: chain
    `-model: sram
      `-Age
        +-n: 1
        `-sum: 0
''', statistics, s, 'bike')
                self.assert_command(f'''item: cotic
+-Age
| +-n: 1
| `-sum: {days('2020-03-24')}
`-component: chain
  +-model: sram
  | `-Age
  |   +-n: 1
  |   `-sum: {days('2020-03-24')}
  +-model: kcm
  | `-Age
  |   +-n: 1
  |   `-sum: 90
  +-model: sram
  | `-Age
  |   +-n: 1
  |   `-sum: 61
  +-model: kcm
  | `-Age
  |   +-n: 1
  |   `-sum: 631
  `-model: sram
    `-Age
      +-n: 1
      `-sum: 30
''', statistics, s, 'cotic')
                self.assert_command(f'''component: chain
+-model: sram
| `-Age
|   +-n: 1
|   `-sum: {days('2020-03-24')}
+-model: kcm
| `-Age
|   +-n: 1
|   `-sum: 90
+-model: sram
| `-Age
|   +-n: 1
|   `-sum: 61
+-model: kcm
| `-Age
|   +-n: 1
|   `-sum: 631
+-model: sram
| `-Age
|   +-n: 1
|   `-sum: 30
`-model: sram
  `-Age
    +-n: 1
    `-sum: 0
''', statistics, s, 'chain')
                self.assert_command(f'''model: sram
`-Age
  +-n: 1
  `-sum: {days('2020-03-24')}
''', statistics, s, 'sram')
                finish(s, 'bowman', None, False)
                with self.assertRaises(Exception) as ctx:
                    finish(s, 'bowman', None, False)
                self.assertTrue('retired' in str(ctx.exception), ctx.exception)
                self.assertEqual(len(KitModel.get_all(s, KitItem.get(s, 'cotic'), KitComponent.get(s, 'chain'))), 5)
                undo(s, 'cotic', 'chain', 'sram', None, True)
                self.assertEqual(len(KitModel.get_all(s, KitItem.get(s, 'cotic'), KitComponent.get(s, 'chain'))), 2)
                undo(s, 'cotic', 'chain', 'kcm', None, True)
                self.assertEqual(len(KitModel.get_all(s, KitItem.get(s, 'cotic'), KitComponent.get(s, 'chain'))), 0)
                undo(s, 'bowman', 'chain', 'sram', None, True)
                self.assertFalse(KitComponent.get(s, 'chain', require=False))

    def assert_command(self, text, cmd, *args):
        args = list(args)
        output = StringIO()
        cmd(*args, output=output)
        self.assertEqual(output.getvalue(), text)

    def test_models(self):
        with TemporaryDirectory() as f:

            args, sys, db = bootstrap_dir(f, m(V), '5', configurator=default)
            args, sys, db = bootstrap_dir(f, m(V), '5', mm(DEV), 'activities',
                                           'data/test/source/personal/2018-08-03-rec.fit',
                                           m(D.upper())+'kit=cotic')
            activities(args, sys, db)
            args, sys, db = bootstrap_dir(f, m(V), '5', mm(DEV), 'activities',
                                           'data/test/source/personal/2018-08-27-rec.fit',
                                           m(D.upper())+'kit=cotic')
            activities(args, sys, db)
            run_pipeline(sys, db, args[BASE], PipelineType.STATISTIC, like=['%Activity%'], n_cpu=1)

            with db.session_context() as s:
                start(s, 'bike', 'cotic', '2018-01-01', True)
                start(s, 'bike', 'marin', '2018-01-01', False)
                change(s, 'cotic', 'chain', 'sram', None, True, True)
                change(s, 'cotic', 'chain', 'kcm', '2018-01-01', False, False)
                change(s, 'cotic', 'chain', 'sram', '2018-05-01', False, False)
                change(s, 'cotic', 'chain', 'kcm', '2018-07-01', False, False)
                change(s, 'cotic', 'chain', 'sram', '2018-04-01', False, False)
                start(s, 'bike', 'bowman', '2018-01-01', False)
                change(s, 'bowman', 'chain', 'sram', None, False, True)

            run_pipeline(sys, db, args[BASE], PipelineType.STATISTIC, like=['%Kit%'], n_cpu=1)

            with db.session_context() as s:
                bike = get_name(s, 'bike').to_model(s, depth=3, statistics=INDIVIDUAL, own_models=False)
                self.assertEqual(bike[TYPE], KitGroup.SIMPLE_NAME)
                self.assertEqual(bike[NAME], 'bike')
                self.assertEqual(len(bike[ITEMS]), 3)
                cotic = [item for item in bike[ITEMS] if item[NAME] == 'cotic'][0]
                self.assertEqual(cotic[TYPE], KitItem.SIMPLE_NAME)
                self.assertEqual(cotic[NAME], 'cotic')
                self.assertEqual(len(cotic[COMPONENTS]), 1)
                chain = cotic[COMPONENTS][0]
                self.assertEqual(chain[TYPE], KitComponent.SIMPLE_NAME)
                self.assertEqual(chain[NAME], 'chain')
                self.assertEqual(len(chain[MODELS]), 6)
                self.assertFalse(STATISTICS in bike)




from logging import getLogger

from . import Reader
from ..calculate.monitor import MonitorCalculator
from ..names import REST_HR, DAILY_STEPS
from ...diary.database import summary_column
from ...diary.model import value, optional_text
from ...sql.tables.statistic import StatisticJournal, StatisticName

log = getLogger(__name__)


class MonitorDiary(Reader):

    @optional_text('Monitor')
    def _read_date(self, s, date):
        for field in self.__read_fields(s, date, DAILY_STEPS, REST_HR):
            yield value(field.statistic_name.name, field.value,
                        units=field.statistic_name.units, measures=field.measures_as_model(date))

    @staticmethod
    def __read_fields(s, date, *names):
        for name in names:
            journal = StatisticJournal.at_date(s, date, name, MonitorCalculator, None)
            if journal:
                yield journal

    @optional_text('Monitor')
    def _read_schedule(self, s, date, schedule):
        for name in self.__names(s, DAILY_STEPS, REST_HR):
            column = list(summary_column(s, schedule, date, name))
            if column:
                yield column

    @staticmethod
    def __names(s, *names):
        for name in names:
            sname = s.query(StatisticName). \
                filter(StatisticName.name == name,
                       StatisticName.owner == MonitorCalculator).one_or_none()
            if sname:
                yield sname




from collections import Counter
from logging import getLogger
from operator import eq, lt, gt
from re import compile
from sys import stdout

from .format.read import filtered_records, filtered_tokens
from .format.records import no_bad_values, fix_degrees, append_units, no_unknown_fields, join_values, \
    to_hex, no_filter, merge_duplicates
from ..commands.args import RECORDS, FIELDS, CSV, TABLES, GREP, TOKENS
from ..lib.io import terminal_width
from ..lib.utils import unique

log = getLogger(__name__)


# this logic now only used in testing - moved to ch2.commands.fit
def summarize(format, data, all_fields=False, all_messages=False, internal=False,
              after_bytes=None, limit_bytes=-1, after_records=None, limit_records=-1,
              messages=None, fields=None, warn=False, profile_path=None, grep=None,
              name_file=None, invert=False, match=1, compact=False, context=False, no_validate=False, max_delta_t=None,
              width=None, output=stdout):

    width = width or terminal_width()

    if (after_records or limit_records != -1) and (after_bytes or limit_bytes != -1):
        raise Exception('Constrain either records or bytes, not both')

    if name_file and format != GREP:
        print()
        print(name_file)

    if format == RECORDS:
        summarize_records(data,
                          all_fields=all_fields, all_messages=all_messages, internal=internal,
                          after_bytes=after_bytes, limit_bytes=limit_bytes,
                          after_records=after_records, limit_records=limit_records,
                          record_names=messages, field_names=fields,
                          warn=warn, no_validate=no_validate, max_delta_t=max_delta_t,
                          profile_path=profile_path, width=width, output=output)
    elif format == TABLES:
        summarize_tables(data,
                         all_fields=all_fields, all_messages=all_messages, internal=internal,
                         after_bytes=after_bytes, limit_bytes=limit_bytes,
                         after_records=after_records, limit_records=limit_records,
                         record_names=messages, field_names=fields,
                         warn=warn, no_validate=no_validate, max_delta_t=max_delta_t,
                         profile_path=profile_path, width=width, output=output)
    elif format == GREP:
        summarize_grep(data, grep,
                       name_file=name_file, match=match, compact=compact, context=context, invert=invert,
                       after_bytes=after_bytes, limit_bytes=limit_bytes,
                       after_records=after_records, limit_records=limit_records,
                       warn=warn, no_validate=no_validate, max_delta_t=max_delta_t, profile_path=profile_path,
                       width=width, output=output)
    elif format == CSV:
        summarize_csv(data,
                      after_bytes=after_bytes, limit_bytes=limit_bytes, internal=internal,
                      after_records=after_records, limit_records=limit_records,
                      record_names=messages, field_names=fields, warn=warn, no_header=no_validate,
                      max_delta_t=max_delta_t, profile_path=profile_path,
                      output=output)
    elif format == TOKENS:
        summarize_tokens(data,
                         after_bytes=after_bytes, limit_bytes=limit_bytes,
                         after_records=after_records, limit_records=limit_records,
                         warn=warn, no_validate=no_validate, max_delta_t=max_delta_t, profile_path=profile_path,
                         output=output)
    elif format == FIELDS:
        summarize_fields(data,
                         after_bytes=after_bytes, limit_bytes=limit_bytes,
                         after_records=after_records, limit_records=limit_records,
                         warn=warn, no_validate=no_validate, max_delta_t=max_delta_t, profile_path=profile_path,
                         output=output)
    else:
        raise Exception('Bad format: %s' % format)


def summarize_tokens(data, after_bytes=None, limit_bytes=-1, after_records=None, limit_records=-1,
                     warn=False, no_validate=False, max_delta_t=None, profile_path=None, output=stdout):

    types, messages, tokens = \
        filtered_tokens(data,
                        after_bytes=after_bytes, limit_bytes=limit_bytes,
                        after_records=after_records, limit_records=limit_records,
                        warn=warn, no_validate=no_validate, max_delta_t=max_delta_t, profile_path=profile_path)

    for index, offset, token in tokens:
        print('%03d %05d %s' % (index, offset, token), file=output)


def summarize_fields(data, after_bytes=None, limit_bytes=-1, after_records=None, limit_records=-1,
                     warn=False, no_validate=False, max_delta_t=None, profile_path=None, output=stdout):

    types, messages, tokens = \
        filtered_tokens(data,
                        after_bytes=after_bytes, limit_bytes=limit_bytes,
                        after_records=after_records, limit_records=limit_records,
                        warn=warn, no_validate=no_validate, max_delta_t=max_delta_t, profile_path=profile_path)

    for index, offset, token in tokens:
        print('%03d %05d %s' % (index, offset, token), file=output)
        for line in token.describe_fields(types):
            print('  %s' % line, file=output)


def summarize_records(data, all_fields=False, all_messages=False, internal=False,
                      after_bytes=None, limit_bytes=-1, after_records=None, limit_records=-1,
                      record_names=None, field_names=None,
                      warn=False, no_validate=False, max_delta_t=None, profile_path=None,
                      width=None, output=stdout):

    types, messages, records = \
        filtered_records(data,
                         after_bytes=after_bytes, limit_bytes=limit_bytes,
                         after_records=after_records, limit_records=limit_records,
                         record_names=record_names, field_names=field_names,
                         warn=warn, no_validate=no_validate, internal=internal,
                         profile_path=profile_path, max_delta_t=max_delta_t, pipeline=[merge_duplicates])

    records = list(records)
    print(file=output)
    pprint_as_dicts(records, all_fields, all_messages, width=width, output=output)


def summarize_tables(data, all_fields=False, all_messages=False, internal=False,
                     after_bytes=None, limit_bytes=-1, after_records=None, limit_records=-1,
                     record_names=None, field_names=None,
                     warn=False, no_validate=False, max_delta_t=None, profile_path=None,
                     width=None, output=stdout):

    types, messages, records = \
        filtered_records(data,
                         after_bytes=after_bytes, limit_bytes=limit_bytes,
                         after_records=after_records, limit_records=limit_records,
                         record_names=record_names, field_names=field_names,
                         warn=warn, no_validate=no_validate, internal=internal,
                         profile_path=profile_path, max_delta_t=max_delta_t, pipeline=[merge_duplicates])

    records = list(record[2] for record in records)
    counts = Counter(record.identity for record in records)
    small, large = partition(records, counts)
    width = width or terminal_width()
    print(file=output)
    pprint_as_dicts(small, all_fields, all_messages, width=width, output=output, full_name=False)
    pprint_as_tuples(large, all_fields, all_messages, width=width, output=output)


class Done(Exception): pass


CMP = compile(r'([^=<>~]+)([=<>~])([^=<>~]+)')


class Matcher:

    def __init__(self, pattern):
        if ':' in pattern:
            record, pattern = pattern.split(':', 1)
            self.record = compile(record).match
        else:
            self.record = None
        match = CMP.match(pattern)
        if match:
            self.field = compile(match.group(1)).match
            if match.group(2) == '~':
                self.value = compile(match.group(2)).match
            else:
                try:
                    try:
                        value = float(match.group(3))
                        log.debug('Matching %f as a float' % value)
                    except ValueError:
                        value = match.group(3)
                        log.debug('Matching %s as a string' % value)
                    self.value = {'=': self.__build_compare(eq, value),
                                  '<': self.__build_compare(lt, value),
                                  '>': self.__build_compare(gt, value)}[match.group(2)]
                except:
                    if match.group(2) != '=':
                        raise Exception('Comparison "%s" with non-numerical value "%s"' %
                                        (match.group(2), match.group(3)))
                    self.value = lambda v: v == match.group(3)
                    print('STRING', self.value)
        else:
            self.field = compile(pattern).match
            self.value = None

    def __build_compare(self, op, value):
        def compare(v):
            if isinstance(value, float):
                try:
                    v = float(v)
                    return op(v, value)
                except:
                    return False
            else:
                return op(str(v), value)
        return compare

    def match(self, record, field, value):
        return (self.record is None or self.record(record)) and \
               (self.field is None or self.field(field)) and \
               (self.value is None or self.value(value))


def summarize_grep(data, grep, name_file=None, match=1, compact=False, context=False, invert=False,
                   after_bytes=None, limit_bytes=-1, after_records=None, limit_records=-1,
                   warn=False, no_validate=False, max_delta_t=None, profile_path=None, width=80, output=stdout):

    types, messages, records = \
        filtered_records(data, warn=warn, no_validate=no_validate, profile_path=profile_path,
                         max_delta_t=max_delta_t, pipeline=[merge_duplicates])
    matchers = [Matcher(pattern) for pattern in grep]
    first, total_matches = True, 0
    first_record = 0 if (after_records is None) else None
    first_bytes = 0 if (after_bytes is None) else None

    try:
        for index, offset, record in records:
            matched_matchers, matched_names_values, display = set(), set(), ''
            if (first_record is None and (after_records is not None and index >= after_records)) or \
                    (first_bytes is None and (after_bytes is not None and offset >= after_bytes)):
                first_record, first_bytes = index, offset
            if first_record is not None or first_bytes is not None:
                if (first_record is not None and (limit_records < 0 or i - first_record < limit_records)) and \
                        (first_bytes is not None and (limit_bytes < 0 or offset - first_bytes < limit_bytes)):
                    record = record.as_dict(fix_degrees, merge_duplicates)
                    for name, values_units in sorted(record.data.items()):
                        if values_units and values_units[0]:
                            for value in values_units[0]:
                                for matcher in matchers:
                                    if matcher.match(record.name, name, value):
                                        matched_matchers.add(matcher)
                                        if (name, value) not in matched_names_values:
                                            display += '%s:%s=%s\n' % (record.name, name, value)
                                            matched_names_values.add((name, value))
                    if len(matched_matchers) == len(matchers):
                        total_matches += 1
                        if match:
                            if first:
                                if context or not compact:
                                    print(file=output)
                                first = False
                            if context:
                                pprint_as_dicts([(index, offset, record)], True, True,
                                                width=width, output=output)
                            else:
                                print(display, file=output, end='' if compact else '\n')
                        if match > -1 and total_matches > max(1, match):
                            raise Done()
                else:
                    raise Done()
    except Done:
        pass
    if name_file:
        if not total_matches == invert:
            print(name_file, file=output)


def summarize_csv(data, after_bytes=0, limit_bytes=-1, after_records=0, limit_records=-1, internal=False,
                  record_names=None, field_names=None, warn=False, no_header=False, max_delta_t=None,
                  profile_path=None, output=stdout):
    types, messages, tokens = \
        filtered_tokens(data,
                        after_bytes=after_bytes, limit_bytes=limit_bytes,
                        after_records=after_records, limit_records=limit_records,
                        warn=warn, no_validate=no_header, max_delta_t=max_delta_t, profile_path=profile_path)
    for index, offset, token in tokens:
        if hasattr(token, 'describe_csv'):
            values = ','.join(str(component)
                              for component in token.describe_csv(warn=warn, internal=internal,
                                                                  record_names=record_names, field_names=field_names))
            if values:
                print(values, file=output)


def partition(records, counts, threshold=3):
    small, large = [], []
    for record in records:
        if counts[record.identity] <= threshold:
            small.append(record)
        else:
            large.append(record)
    return small, large


def pprint_as_dicts(records, all_fields, all_messages, width=80, output=stdout, full_name=True):
    for record in records:
        if full_name:
            index, offset, record = record
        if all_messages or record.is_known():
            record = record.as_dict(join_values, append_units, to_hex, fix_degrees,
                                    no_filter if all_fields else no_unknown_fields,
                                    no_bad_values)
            if full_name:
                print('%03d %05d %s' % (index, offset, record.identity), file=output)
            else:
                print(record.name, file=output)
            pprint_with_tabs(('%s: %s' % (name, value) for name, value in sorted(record.data.items())),
                             width=width, output=output)
            print(file=output)


def sort_names(data):
    return sorted(list(data), key=lambda x: ' ' if x[0] == 'timestamp' else x[0])


def pprint_as_tuples(records, all_fields, all_messages, width=80, output=stdout):
    records = [record.as_dict(sort_names, timestamp=((record.timestamp,), 's'))
               for record in records]
    titles = [record.as_names(no_filter if all_fields else no_unknown_fields)
              for record in unique(records, key=lambda x: x.identity)
              if all_messages or record.is_known()]
    for title in titles:
        pprint_series(title,
                      [record.as_values(join_values, append_units, to_hex, fix_degrees,
                                        no_filter if all_fields else no_unknown_fields)
                       for record in records
                       if record.identity == title.identity],
                      width=width, output=output)


def measure_lengths(records, lengths=None, separator=',', keep_bad=False):
    for record in records:
        if lengths is None:
            lengths = [(len(datum) + (len(separator) if i+1 < len(record.data) else 0)) if datum else 0
                       for i, datum in enumerate(record.data)]
        else:
            for i, datum in enumerate(record.data):
                if datum and (lengths[i] or not keep_bad):
                    lengths[i] = max(lengths[i], len(datum) + (len(separator) if i+1 < len(record.data) else 0))
    return lengths


def pad_to_lengths(record, lengths, separator=',', bad='-'):
    for i, datum in enumerate(record.data):
        if lengths[i]:
            if datum:
                yield (datum + (separator if i+1 < len(lengths) else '')).ljust(lengths[i])
            else:
                yield bad.ljust(lengths[i])


def pprint_series(title, records, width=80, output=stdout):
    print(title.identity, file=output)
    lengths = measure_lengths([title], keep_bad=True, lengths=measure_lengths(records))
    pprint_with_tabs(pad_to_lengths(title, lengths), first_indent=2, indent=4, separator='',
                     width=width, output=output)
    for record in records:
        pprint_with_tabs(pad_to_lengths(record, lengths), first_indent=2, indent=4, separator='',
                         width=width, output=output)
    print(file=output)


def pprint_with_tabs(data, first_indent=None, indent=2, tab=4, width=80, min_space=2, separator=',', output=stdout):
    if first_indent is None:
        first_indent = indent
    min_space -= 1
    line = ' ' * first_indent
    first = True
    for datum in data:
        if not first:
            line += separator
        l_old = len(line)
        if l_old > indent:
            l_new = (1 + (l_old - indent + min_space) // tab) * tab + indent
            # equality here leaves an attractive space at end of line
            if l_new + len(datum) >= width:
                print(line, file=output)
                line = ' ' * indent + datum
            else:
                line += ' ' * (l_new - l_old) + datum
        else:
            line += datum
        first = False
    if len(line) > indent:
        print(line, file=output)




from distutils.util import strtobool

import numpy as np
import scipy as sp
import pandas as pd
import matplotlib.pyplot as plt
from bokeh.plotting import figure, output_notebook, show
from scipy.optimize import fmin

from ch2.data.power import *
from ch2.lib import bookend
from ch2.jupyter.decorator import template


@template
def fit_power_parameters(bookmark, large):

    f'''
    # Fit Power Parameters to {bookmark}

    This notebook allows you to calculate power parameters (CdA - air resistance, Crr - rolling resistance)
    from bookmarked activities.

    Beforehand you should have generated the bookmark by running

        > python -m ch2.data.coasting

    or similar to identify sections of activities with little pedalling.
    See that module for more information.

    The `large` parameter means that each bookmark is taken in its entirety.
    The alternative is that they are divided into small sub-samples reflecting the data sample rate.
    '''

    '''
    $contents
    '''

    '''
    ## Load Data
    
    Open a connection to the database and load the data we require.
    '''
    s = session('-v 5')
    large = strtobool(large)
    route = activity_statistics(s, LATITUDE, LONGITUDE, SPHERICAL_MERCATOR_X, SPHERICAL_MERCATOR_Y, DISTANCE,
                                ELEVATION, SPEED, CADENCE, bookmarks=bookmarks(s, bookmark))
    route.sort_index(inplace=True)  # bookmarks are not sorted by time
    if large:
        route, max_gap = bookend(route), None
    else:
        max_gap = 10
    route = add_differentials(route, max_gap=max_gap)
    if large:
        route = route.iloc[1::2]
    route.describe()

    '''
    ## Add Energy Calculations
    '''
    weight = 64+12  # weight of rider + bike / kg  todo - extract weight from db
    route = add_energy_budget(route, weight)
    route = add_air_speed(route)
    route = add_cda_estimate(route)
    route = add_crr_estimate(route, weight)
    route.describe()

    '''
    ## Plot Constraints
    
    The calculations above added an energy budget for each "step" in the data.
    These included values for CdA and Crr that would "explain" the decrease in energy 
    (taking each alone - so the CdA is that required for all energy lost to air resistance, 
    the Crr is that required for all energy lost to rolling resistance).
    
    But we know that both CdA and Crr could be important.
    So what we want is a linear combination of the two.
    For example, maybe the energy loss is 90% due to CdA and 10% due to Crr.
    All these possible linear combinations lie on a line that joins 100% CdA and 0% Crr with 0% CdA and 100% Crr.
    
    So the plot below shows all possible combinations of CdA and Crr.
    And what we are looking for is the most common value.
    So we want to know where the plot is darkest / the lines are most dense. 
    '''
    output_notebook()
    f = figure(plot_width=500, plot_height=400)
    clean = route.loc[route[DELTA_ENERGY] < 0].dropna()
    cs = pd.DataFrame({CDA: [(0, cda) for cda in clean[CDA]],
                       CRR: [(crr, 0) for crr in clean[CRR]]})
    f.multi_line(xs=CDA, ys=CRR, source=cs, line_alpha=0.1, line_color='black')
    f.xaxis.axis_label = 'CdA'
    f.yaxis.axis_label = 'Crr'
    show(f)

    '''
    ## CdA Only
    
    If we ignore Crr then we can estimate CdA by looking at the relative number of constraints of CdA
    where Crr is zero.
    
    We do this by fitting to binned data.
    The peak in the fit(s) gives the value of CdA if Crr is unimportant.
    '''
    bins = np.linspace(0, 1.5, 30)
    width = bins[1] - bins[0]
    counts = clean[CDA].groupby(pd.cut(clean[CDA], bins)).size()
    print(counts.describe())

    cda = pd.DataFrame({CDA: 0.5 * (bins[:-1] + bins[1:]), 'n': counts.values})
    f = figure(plot_width=900, plot_height=300)
    f.quad(top=counts, left=bins[:-1]+0.1*width, right=bins[1:]-0.1*width, bottom=0)
    for order in range(2, 20, 2):
        coeff = sp.polyfit(cda[CDA], cda['n'], order)
        p = sp.poly1d(coeff)
        print(order, fmin(lambda x: -p(x), 0.6, disp=0)[0])
        f.line(x=cda['CdA'], y=p(cda['CdA']), line_color='orange')
    show(f)

    '''
    ## Sample Constraints
    
    If we want to include Crr then we need to find a way to measure the "peak" in the messy line plot above.
    To do this we convert to a collection of points and then fit a 2D density function.
    
    Conversion to points is done by selecting points at random on each line.
    (You might think that shorter lines should generate less points.
    The way I see it, each line is an observation that constrains CdA and Crr.
    Each observation has equal weight, so each line generates a point.)
    
    Random points avoids any systematic patterns from uniform sampling 
    and allows re-runs to give some idea of noise. 
    '''

    def sample():
        clean.loc[:, 'random'] = np.random.random(size=len(clean))
        clean.loc[:, 'x'] = clean[CDA] * clean['random']
        clean.loc[:, 'y'] = clean[CRR] * (1 - clean['random'])
        return clean.loc[:, ['x', 'y']]

    s = pd.concat([sample() for _ in range(100 if large else 10)])
    print(s.describe())

    f = figure(plot_width=600, plot_height=600)
    f.scatter(x='x', y='y', source=s)
    show(f)

    '''
    ## Smooth, Find Maximum
    
    We generate and plot a Gaussian kernel density estimate.
    
    See https://towardsdatascience.com/simple-example-of-2d-density-plots-in-python-83b83b934f67
    
    You may want to play around with bandwidth by supplying a second argument to gaussian_kde. 
    See https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.gaussian_kde.html
    '''

    kernel = sp.stats.gaussian_kde(s.transpose())

    xmin, xmax = 0, 1
    ymin, ymax = 0, 0.02
    xx, yy = np.mgrid[xmin:xmax:100j, ymin:ymax:100j]
    xy = np.vstack([xx.ravel(), yy.ravel()])
    smooth = np.reshape(kernel(xy), xx.shape)

    fig = plt.figure(figsize=(8,8))
    ax = fig.gca()
    ax.set_xlim(xmin, xmax)
    ax.set_ylim(ymin, ymax)
    ax.contourf(xx, yy, smooth, cmap='coolwarm')
    cset = ax.contour(xx, yy, smooth, colors='k')
    ax.clabel(cset, inline=1, fontsize=10)
    ax.set_xlabel('CdA')
    ax.set_ylabel('Crr')
    plt.title('2D Gaussian Kernel density estimation')

    '''
    For my data this shows (roughly):
    
    * large=True: Crr ~ 0.005-0.006 and CdA ~ 0.40-0.45
    
    * large=False: Crr ~ 0.01 and CdA ~ 0.35-0.40
    
    which supports the idea that Crr isn't well-constrained by the data.
    
    I chose large=True values since I suspect the larger intervals make the elevation values more accurate, 
    
    Since I am loading data with a command like
    
        > ch2 activities -D kit=cotic -- ~/archive/fit/**/*.fit
    
    I define this constant:
    
        > ch2 constants add --single Power.cotic \
            --description 'Bike namedtuple values to calculate power for this kit' \
            --validate ch2.stats.calculate.power.Bike
        > ch2 constants set Power.cotic '{"cda": 0.42, "crr": 0.0055, "weight": 12}'
        
    With that, the standard configuration should calculate power estimates.
    '''




from logging import getLogger

from ..config import Config

log = getLogger(__name__)


def default(sys, s, base, no_diary=False):
    '''
## default

The default configuration with basic activity groups, diary topics and FF parameters.
    '''
    Config(sys, base, no_diary=no_diary).load(s)




from logging import getLogger
from tempfile import TemporaryDirectory
from tests import LogTestCase

from sqlalchemy.sql.functions import count

from ch2.commands.args import bootstrap_dir, m, V, DEV, mm, FORCE
from ch2.commands.constants import constants
from ch2.config.profile.default import default
from ch2.sql.tables.constant import Constant

log = getLogger(__name__)


class TestConstant(LogTestCase):

    def test_constant(self):
        with TemporaryDirectory() as f:
            args, sys, db = bootstrap_dir(f, m(V), '5')
            bootstrap_dir(f, m(V), '5', mm(DEV), configurator=default)
            with db.session_context() as s:
                n = s.query(count(Constant.id)).scalar()
                self.assertEqual(n, 15)
            args, sys, db = bootstrap_dir(f, m(V), '5', 'constants', 'set', 'FTHR.%', '154', mm(FORCE))
            constants(args, sys, db)
            with db.session_context() as s:
                n = s.query(count(Constant.id)).scalar()
                self.assertEqual(n, 15)
                # todo - maybe test for value?
                # todo - now that this is defined anyway, change the test?



from os.path import join

from ch2.commands.args import RECORDS
from ch2.fit.fix import fix
from ch2.fit.format.tokens import FileHeader
from ch2.fit.profile.profile import read_fit
from ch2.fit.summary import summarize
from ch2.lib.tests import OutputMixin
from tests import LogTestCase


class TestFixFit(LogTestCase, OutputMixin):

    def setUp(self):
        self.test_dir = 'data/test'
        self.profile_path = 'data/sdk/Profile.xlsx'

    def test_null(self):
        good = read_fit(join(self.test_dir, 'source/personal/2018-08-27-rec.fit'))
        same = fix(bytearray(good))
        self.assertTrue(good is not same)  # check making a copy
        self.assertEqual(good, same)

    def test_null_drop(self):
        good = read_fit(join(self.test_dir, 'source/personal/2018-08-27-rec.fit'))
        same = fix(bytearray(good), drop=True, fix_checksum=True, fix_header=True)
        self.assertTrue(good is not same)  # check making a copy
        self.assertEqual(good, same)

    def test_null_slices(self):
        good = read_fit(join(self.test_dir, 'source/personal/2018-08-27-rec.fit'))
        same = fix(bytearray(good), slices=':', fix_checksum=True, fix_header=True)
        self.assertTrue(good is not same)  # check making a copy
        self.assertEqual(good, same)

    def test_drop(self):
        bad = read_fit(join(self.test_dir, 'source/other/8CS90646.FIT'))
        fixed = fix(bad, drop=True, fix_checksum=True, fix_header=True)
        self.assertTrue(len(fixed) < len(bad))
        with self.assertTextMatch('data/test/target/other/TestFixFit.test_drop') as output:
            summarize(RECORDS, fixed, output=output)

    def test_slices(self):
        bad = read_fit(join(self.test_dir, 'source/other/8CS90646.FIT'))
        with self.assertRaisesRegex(Exception, 'Error fixing checksum'):
            fix(bad, slices=':1000', fix_checksum=True, fix_header=True)  # first 1k bytes only

    def test_no_last_byte(self):
        good = read_fit(join(self.test_dir, 'source/personal/2018-08-27-rec.fit'))
        same = fix(bytearray(good), drop=True, fix_checksum=True, fix_header=True)
        self.assertEqual(same, good)
        fixed = fix(bytearray(good)[:-1], drop=True, fix_checksum=True, fix_header=True)
        self.assertEqual(fixed, good)
        fixed = fix(bytearray(good)[:-2], drop=True, fix_checksum=True, fix_header=True)
        self.assertEqual(fixed, good)

    def test_no_header(self):
        good = read_fit(join(self.test_dir, 'source/personal/2018-08-27-rec.fit'))
        same = fix(bytearray(good), drop=True, fix_checksum=True)
        self.assertTrue(good is not same)  # check making a copy
        self.assertEqual(same, good)
        header = FileHeader(good)
        with self.assertRaisesRegex(Exception, 'Error fixing checksum'):
            fix(bytearray(good)[len(header):], fix_checksum=True, fix_header=True)
        fixed = fix(bytearray(good)[len(header):],
                    add_header=True, drop=True, fix_checksum=True, fix_header=True)
        self.assertEqual(good, fixed)
        fixed = fix(bytearray(good), add_header=True, slices=':14,28:', fix_checksum=True, fix_header=True)
        self.assertEqual(good, fixed)

    def test_other_header(self):
        bad = read_fit(join(self.test_dir, 'source/other/8CS90646.FIT'))
        old_header = FileHeader(bad)
        fixed = fix(bytearray(bad), drop=True, header_size=27, fix_checksum=True, fix_header=True)
        new_header = FileHeader(fixed)
        self.assertEqual(new_header.header_size, 27)
        self.assertEqual(new_header.protocol_version, old_header.protocol_version)
        self.assertEqual(new_header.profile_version, old_header.profile_version)

    def test_other_bad(self):
        bad = read_fit(join(self.test_dir, 'source/other/2018-04-15-09-18-20.fit'))
        fixed = fix(bytearray(bad), drop=True, fix_checksum=True, fix_header=True)
        with self.assertTextMatch(join(self.test_dir, 'target/other/TestFixFit.test_unknown_bad:1')) as output:
            summarize(RECORDS, fixed, output=output)
        bad = read_fit(join(self.test_dir, 'source/other/2018-02-24-10-04-10.fit'))
        fixed = fix(bytearray(bad), drop=True, fix_checksum=True, fix_header=True)
        with self.assertTextMatch(join(self.test_dir, 'target/other/TestFixFit.test_unknown_bad:2')) as output:
            summarize(RECORDS, fixed, output=output)

    def test_other_good(self):
        good = read_fit(join(self.test_dir, 'source/other/77F73023.FIT'))
        same = fix(bytearray(good))
        self.assertEqual(good, same)
        self.assertFalse(good is same)

    def test_pyfitparse_fix_header(self):
        for file in ('activity-filecrc.fit',  # bad checksum
                     'activity-activity-filecrc.fit',  # data size incorrect
                     'activity-settings.fit',  # data size incorrect
                     ):
            bad = read_fit(join(self.test_dir, 'source/python-fitparse', file))
            with self.assertBinaryMatch(join(self.test_dir, 'source/python-fitparse-fix', file)) as output:
                output.write(fix(bad, fix_checksum=True, fix_header=True))

    def test_pyfitparse_fix_drop(self):
        for file in ('activity-unexpected-eof.fit',  # data size incorrect
                     'activity-settings-nodata.fit',   # data size incorrect
                     'activity-settings-corruptheader.fit',  # data size incorrect
                     ):
            bad = read_fit(join(self.test_dir, 'source/python-fitparse', file))
            with self.assertBinaryMatch(join(self.test_dir, 'source/python-fitparse-fix', file)) as output:
                output.write(fix(bad, drop=True, fix_checksum=True, fix_header=True))

    def test_pyfitparse_fix_drop_2(self):
        for file in ('event_timestamp.fit',  # data size incorrect
                     'antfs-dump.63.fit',  # strange timestamp
                     'compressed-speed-distance.fit',  # strange timestamp
                     ):
            bad = read_fit(join(self.test_dir, 'source/python-fitparse', file))
            with self.assertBinaryMatch(join(self.test_dir, 'source/python-fitparse-fix', file)) as output:
                output.write(fix(bad, drop=True, max_drop_cnt=2, fix_checksum=True, fix_header=True))

    def test_drop_bug(self):
        bad = read_fit(join(self.test_dir, 'source/personal/2018-07-26-rec.fit'))
        fix(bad, drop=True, fix_checksum=True, fix_header=True, max_delta_t=60, max_fwd_len=500)




from logging import getLogger

from . import MultiProcCalculator, ActivityJournalCalculatorMixin
from ...sql import StatisticJournal, Timestamp
from ...sql.tables.kit import expand_item

log = getLogger(__name__)


class KitCalculator(ActivityJournalCalculatorMixin, MultiProcCalculator):
    '''
    Convert `-D kit=XXX` statistic set on activities during import into kit usage entries.

    The kit statistic is also used directly by other code (eg to filter activities by kit).
    '''

    def _run_one(self, s, time):
        ajournal = self._get_source(s, time)
        with Timestamp(owner=self.owner_out, source=ajournal).on_success(s):
            kit = StatisticJournal.for_source(s, ajournal.id, 'kit', self.owner_in, ajournal.activity_group)
            if kit:
                log.debug(f'Read {kit.value} at {time} / {ajournal.activity_group.name}')
                for kit_name in kit.value.split(','):
                    try:
                        for kit_instance in expand_item(s, kit_name, ajournal.start):
                            kit_instance.add_use(s, ajournal.start, source=ajournal, owner=self.owner_out)
                            log.debug(f'Added usage for {kit_instance}')
                    except Exception as e:
                        log.warning(f'Could not add statistics for {kit_name}: {e}')
            else:
                log.debug(f'No kit defined for this activity ({time})')



from abc import abstractmethod
from csv import reader
from io import StringIO, BytesIO
from contextlib import AbstractContextManager
from itertools import zip_longest
from logging import getLogger
from os import makedirs
from os.path import exists, dirname, split, join
from re import sub
from tempfile import NamedTemporaryFile

from .utils import grouper

HEX_ADDRESS = lambda s: sub(r'0x[0-9a-f]{8,}', 'ADDRESS', s)


def EXC_HDR_CHK(data):
    for row in data:
        if row[0] in ('FileHeader', 'Checksum'):
            pass
        elif row[0] == 'DeveloperField':
            yield ['Data'] + row[1:]
        else:
            yield row


def EXC_FLD(name):
    def filter(data):
        for row in data:
            try:
                i = row.index(name)
                if not i % 3:
                    row = row[:i] + row[i+3:]
            except ValueError:
                pass
            yield row
    return filter


def EXC_MSG(name):
    def filter(data):
        for row in data:
            if row[0] == 'Data' and row[2] == name:
                yield row[:3]
            else:
                yield row
    return filter


def RNM_UNKNOWN(data):
    for row in data:
        for i in range(0, len(row), 3):
            if row[i].startswith('@'):
                row[i] = 'unknown'
        for i in range(2, len(row), 3):
            if row[i].startswith('MESSAGE'):
                row[i] = 'unknown'
        yield row


def ROUND_DISTANCE(data):
    for row in data:
        for i in range(0, len(row), 3):
            if row[i] == 'distance' and row[i+1] and '.' in row[i+1] and len(row[i+1].split('.')[1]) > 2:
                row[i+1] = '%.2f' % (float(row[i+1]) + 0.0000001)  # round up
        yield row


def sub_extn(path, extn):
    dir, file = split(path)
    return join(dir, '%s.%s' % (file.rsplit('.', 1)[0], extn))


def sub_dir(path, new_dir, offset):
    a, b = split(path)
    if offset:
        return join(sub_dir(a, new_dir, offset-1), b)
    else:
        return join(a, new_dir)


class BaseBufferContext(AbstractContextManager):

    def __init__(self, log, path, filters):
        self._log = log
        self._path = path
        self._filters = filters
        self._buffer = self._make_buffer()

    @abstractmethod
    def _make_buffer(self):
        raise NotImplementedError()

    def __enter__(self):
        return self._buffer

    def _filter(self, data):
        if self._filters:
            for filter in self._filters:
                data = filter(data)
        return data


class TextBufferContext(BaseBufferContext):

    def _make_buffer(self):
        return StringIO()


class BinaryBufferContext(BaseBufferContext):

    def _make_buffer(self):
        return BytesIO()


class TextEqualContext(TextBufferContext):

    def __init__(self, log, path, filters, test):
        super().__init__(log, path, filters)
        self._test = test

    def __exit__(self, exc_type, exc_val, exc_tb):
        if not exc_type:
            with open(self._path, 'r') as input:
                target = input.read()
                result = self._filter(self._buffer.getvalue())
                if result != target:
                    with NamedTemporaryFile(delete=False) as f:
                        f.write(result.encode())
                        self._log.warning('Wrote copy of result to %s' % f.name)
                        self._log.warning('Comparing with %s' % self._path)
                        self._log.warning('diff %s %s' % (f.name, self._path))
                self._test.assertEqual(result, target)


class BinaryEqualContext(BinaryBufferContext):

    def __init__(self, log, path, filters, test):
        super().__init__(log, path, filters)
        self._test = test

    def __exit__(self, exc_type, exc_val, exc_tb):
        if not exc_type:
            with open(self._path, 'rb') as input:
                target = input.read()
                result = self._filter(self._buffer.getvalue())
                if result != target:
                    with NamedTemporaryFile(delete=False) as f:
                        f.write(result)
                        self._log.warning('Wrote copy of result to %s' % f.name)
                        self._log.warning('Comparing with %s' % self._path)
                        self._log.warning('cmp -l %s %s' % (f.name, self._path))
                self._test.assertEqual(result, target)


class TextDumpContext(TextBufferContext):

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type:
            self._log.warning('Error writing %s' % self._path)
        else:
            makedirs(dirname(self._path), exist_ok=True)
            with open(self._path, 'w') as output:
                output.write(self._filter(self._buffer.getvalue()))
            self._log.warning('Wrote data to "%s"' % self._path)


class BinaryDumpContext(BinaryBufferContext):

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type:
            self._log.warning('Error writing %s' % self._path)
        else:
            makedirs(dirname(self._path), exist_ok=True)
            with open(self._path, 'wb') as output:
                output.write(self._buffer.getvalue())
            self._log.warning('Wrote data to "%s"' % self._path)


class CSVEqualContext(TextBufferContext):

    def __init__(self, log, path, filters, test):
        super().__init__(log, path, filters)
        self._test = test

    def __exit__(self, exc_type, exc_val, exc_tb):
        if not exc_type:
            with open(self._path, 'r') as them_in:
                them_reader = self._filter(reader(them_in))
                next(them_reader)  # drop titles
                us_data = self._buffer.getvalue()
                us_reader = self._filter(reader(us_data.splitlines()))
                for row, (us_row, them_row) in enumerate(zip_longest(us_reader, them_reader)):
                    self.compare_rows(row, us_row, them_row, us_data)

    def build_dict(self, row):
        return dict((name, (value, units)) for name, value, units in grouper(row, 3))

    def compare_rows(self, row, us_row, them_row, us_data):
        if us_row[0] == 'CompressedTimestamp': us_row[0] = 'Data'
        self._test.assertEqual(us_row[:3], them_row[:3], 'Row %d header' % row)
        us_dict = self.build_dict(us_row[3:])
        them_dict = self.build_dict(them_row[3:])
        keys = set(us_dict.keys()).union(them_dict.keys())
        for key in keys:
            self.compare_key(key, us_dict, them_dict, row, us_data)

    def compare_key(self, key, us_dict, them_dict, row, us_data):
        if key and key != 'unknown':
            us_value, us_units = us_dict.get(key, ('', ''))
            them_value, them_units = them_dict.get(key, ('', ''))
            try:
                if us_value != 'COMPOSITE':
                    self._test.assertEqual(self.fix_value(us_value), self.fix_value(them_value),
                                           'Value for %s row %d' % (key, row))
                    if us_value:  # avoid too many errors
                        self._test.assertEqual(us_units, them_units, 'Units for %s row %d' % (key, row))
            except Exception:
                self.dump(row, us_data)
                raise

    def fix_value(self, value):
        if value:
            if value.endswith('.0'): value = value[:-2]
            if value == 'False': value = '0'
            if value == 'True': value = '1'
        return value

    def dump(self, row, us_data):
        with NamedTemporaryFile(delete=False) as f:
            f.write(us_data.encode())
            self._log.warning('Wrote copy of CSV to %s' % f.name)
            self._log.warning('Comparing with %s' % self._path)
            self._log.warning('head -n %d %s | tail -n 1; head -n %d %s | tail -n 1' %
                           (row+2, f.name, row+2, self._path))


class OutputMixin:

    def _resolve_log(self, log):
        '''
        Allow mixin with classes that may have their own log.
        '''
        if hasattr(self, 'log'):
            log = log or self.log
        if hasattr(self, '_log'):
            log = log or self._log
        return log or getLogger()

    def assertTextMatch(self, path, log=None, filters=None):
        log = self._resolve_log(log)
        if exists(path):
            return TextEqualContext(log, path, filters, self)
        else:
            log.warning('File "%s" does not exist; will generate rather than check' % path)
            return TextDumpContext(log, path, filters)

    def assertBinaryMatch(self, path, log=None, filters=None):
        log = self._resolve_log(log)
        if exists(path):
            return BinaryEqualContext(log, path, filters, self)
        else:
            log.warning('File "%s" does not exist; will generate rather than check' % path)
            return BinaryDumpContext(log, path, filters)

    def assertCSVMatch(self, path, log=None, filters=None):
        log = self._resolve_log(log)
        if exists(path):
            return CSVEqualContext(log, path, filters, self)
        else:
            # the CSV file should have been generated using the SDK (dev/build-csv.sh)
            raise Exception('Could not open %s' % path)
